from io import BytesIO

import pytest
from aext_panels_server.schemas import (
    MetaResponse,
    PAProjectInfo,
    ProjectInfoData,
    RequestError,
    ServiceResponse,
    VersioningData,
)


def test_pa_project_info_to_bytes_io():
    project_info = PAProjectInfo(
        title="Project",
        project_id="id",
        notebook_file_name="filename",
        notebook_file_size="filesize",
        environment="environment",
        path="path",
    )

    assert type(project_info.to_bytes_io()) == BytesIO


@pytest.mark.parametrize("status, expected", [("", False), ("published", True)])
def test_project_info_data_is_active(status, expected):
    projeect_info_data = ProjectInfoData(
        title="title",
        project_id="project_id",
        versioning=VersioningData(next_version="2023.1.1"),
        notebook_file_name="filename",
        notebook_file_size="filesize",
        environment="environment",
        permission="permission",
        status=status,
    )
    assert projeect_info_data.is_active() == expected


@pytest.mark.parametrize("status, expected", [(True, False), (False, True)])
def test_service_response_has_failed_has_succeeded(status, expected):
    service_response = ServiceResponse(
        meta=MetaResponse(
            error=RequestError(status=status),
        ),
        content="",
    )
    has_failed_expected = not expected
    assert service_response.has_succeeded() == expected
    assert service_response.has_failed() == has_failed_expected


def test_service_response_set_error_message():
    service_response = ServiceResponse(
        meta=MetaResponse(
            error=RequestError(status=True),
        ),
        content="",
    )
    assert service_response.get_error_message() == ""
    service_response.set_error_message("error")
    assert service_response.get_error_message() == "error"


def test_service_response_set_error_status():
    service_response = ServiceResponse(
        meta=MetaResponse(
            error=RequestError(status=True),
        ),
        content="",
    )
    assert service_response.has_failed()
    service_response.set_error_status(False)
    assert service_response.has_succeeded()



from __future__ import annotations

import json
from typing import Dict, Type, TypedDict
from uuid import UUID

import httpx
import tornado
from tornado.routing import _RuleList

from aext_shared.backend_proxy import backend_proxy
from aext_shared.config import get_config
from aext_shared.handler import BackendHandler, create_rules

from .base_handlers import NotebookBaseHandler, PABaseHandler
from .const import MS_API_Version
from .exceptions import (
    CantCreateProject,
    CantCreateProjectVersion,
    CantUpdateProject,
    CondaEnvironmentNotFound,
)
from .handlers_helpers import create_response
from .logger import custom_logger
from .request_schemas.notebook_projects_microservice import (
    CreateProject,
    CreateProjectVersion,
    UpdateProject,
)
from .schemas import (
    MetaResponse,
    ProjectInfoData,
    RequestError,
    ServiceResponse,
    VersioningData,
)
from .services.deployment_microservice import project_microservice_service
from .services.panel import panel_service
from .services.project import project_service
from .utils import to_json

config = get_config()


class ProjectInfo(TypedDict):
    name: str
    path: str
    description: str
    versions: list[str]


class HealthzRouteHandler(BackendHandler):
    @tornado.web.authenticated
    async def get(self):
        self.finish({"live": True})


class ProjectInfoHandler(NotebookBaseHandler):
    """
    Return unique CalVar version for a project by querying existing
    versions and incrementing until a new version is found.
    """

    @tornado.web.authenticated
    async def get(self):
        self.set_header("Content-Type", "application/json")
        file_path = self.get_argument("notebook_file_path", None)
        project_id = self.get_argument("project_id", None)

        if not file_path.endswith(".ipynb"):
            file_path += ".ipynb"
        file_info = await self._get_file_info(f"{file_path}")
        if not file_info:
            response = create_response(status=True, message="Project file_path not found")
            return self.finish(response, status_code=400)

        pa_project_info = None
        try:
            # Check if project already exists
            if project_id:
                pa_project_info = await project_service.get_project_info(self.api, project_id)
        except Exception:
            custom_logger.error("Not possible to get the project from PythonAnywhere")

        # if not exists a pa_project_info, create a new one project
        if pa_project_info is None:
            # Get Conda environment
            try:
                environment = await self.extract_notebook_environment(file_info.file_name)
            except CondaEnvironmentNotFound:
                err_msg = "Notebook doesn't have a Conda environment"
                response = create_response(status=True, message=err_msg, content="")
                return self.finish(response, status_code=400)

            new_project = await project_service.create_project(file_info, environment)
            response = create_response(message="", content=new_project)
        else:
            next_version = project_service._create_or_get_version(pa_project_info.versions[-1])
            versioning = VersioningData(
                next_version=next_version,
                current_version=pa_project_info.versions[-1],
                all_versions=pa_project_info.versions,
            )
            existing_project = ProjectInfoData(
                title=pa_project_info.title,
                project_id=pa_project_info.project_id,
                versioning=versioning,
                notebook_file_name=pa_project_info.notebook_file_name,
                notebook_file_size=pa_project_info.notebook_file_size,
                permission="base",
                environment=pa_project_info.environment,
                status="",
                description=pa_project_info.description,
                domains=[],
            )
            response = create_response(message="", content=existing_project)

        return self.finish(response)


class ProjectHandler(NotebookBaseHandler):
    @tornado.web.authenticated
    async def post(self):
        self.set_header("Content-Type", "application/json")
        path = self.get_argument("path")
        title = self.get_argument("title")
        version = self.get_argument("version")
        description = self.get_argument("description", default="")
        environment = self.get_argument("environment", default="base")
        project_id = self.get_argument("project_id", default=None)

        notebook_path = self._get_path(path)
        notebook_file_info = await self._get_file_info(notebook_path)
        user_access = await self.get_user_access_credentials()

        # Validate Project version
        if not project_service.is_valid_version(version):
            response = create_response(status=True, message="Project Version is not valid", content="")
            return self.finish(response, status_code=400)

        # Validate notebook file path
        if not notebook_file_info:
            response = create_response(status=True, message="Project file_path not found", content="")
            return self.finish(response, status_code=400)

        try:
            await self._log_notebook_data(notebook_path, "aext-cloud-panels-notebook-data")
        except Exception as ex:
            custom_logger.error(f"Error while trying to log: {vars(ex)}")

        # Get Conda environment
        try:
            environment = await self.extract_notebook_environment(notebook_path)
        except CondaEnvironmentNotFound:
            response = create_response(status=True, message="Conda environment not found", content="")
            return self.finish(response, status_code=400)
        err = RequestError(status=False)
        custom_logger.info("Creating project info")
        project_info = await project_service.create_project_info(
            title,
            version,
            description,
            notebook_file_info,
            environment,
        )
        custom_logger.info("Generating project files")
        project_content = await project_service.generate_files(
            user_access.username,
            title,
            version,
            project_info,
            environment,
            notebook_path,
        )
        try:
            api_version = self.request.headers.get("X-MS-API-Version", MS_API_Version.V1)
            user_access = await self.get_user_access_credentials()
            async with project_microservice_service.access_credentials(user_access):
                async with project_microservice_service.incoming_request(self.request):
                    if not project_id:
                        custom_logger.info("Creating a new project")
                        request_data = CreateProject(
                            user_id=user_access.username,
                            title=title,
                            version=version,
                            environment=environment,
                            notebook_file_name=notebook_file_info.file_name,
                            notebook_file_size=notebook_file_info.file_size,
                            permission="public",
                            description=description,
                        )
                        project_response = await project_microservice_service.create_project(
                            request_data, project_content.archive_content, api_version
                        )
                    else:
                        custom_logger.info("Updating the project")
                        project_response = await project_microservice_service.update_project(
                            project_id,
                            UpdateProject(title=title, description=description),
                            project_content.archive_content,
                        )
        except CantCreateProject:
            project_response = "Can't create new project"
            err.status = True
            err.message = "Failure while creating new project"
        except CantUpdateProject:
            project_response = "Can't create new project"
            err.status = True
            err.message = "Failure while updating project"

        response = to_json(ServiceResponse(meta=MetaResponse(error=err), content=project_response))
        self.finish(response)


class ProjectDownloadHandler(PABaseHandler):
    @tornado.web.authenticated
    async def get(self):
        project_id = self.get_argument("project_id")
        project_version = self.get_argument("project_version")

        if not project_id or not project_version:
            return self.finish(
                to_json(
                    ServiceResponse(
                        content=None,
                        meta=MetaResponse(RequestError(True, "Missing project_id and/or project_version")),
                    )
                ),
                status_code=httpx.codes.BAD_REQUEST,
            )

        self.set_header("Content-Type", "application/json, charset=utf-8")
        api_version = self.request.headers.get("X-MS-API-Version", MS_API_Version.V1)
        user_access = await self.get_user_access_credentials()
        async with project_microservice_service.access_credentials(user_access):
            async with project_microservice_service.incoming_request(self.request):
                response = await project_microservice_service.download_project_files(
                    project_id, project_version, api_version
                )
        return self.finish(to_json(response))


class DeploymentHandler(PABaseHandler):
    @tornado.web.authenticated
    async def get(self):
        self.set_header("Content-Type", "application/json")
        user_access = await self.get_user_access_credentials()
        async with project_microservice_service.access_credentials(user_access):
            async with project_microservice_service.incoming_request(self.request):
                custom_logger.debug("Calling microservice to list projects")
                response = await project_microservice_service.list_projects(user_access.username)
        return self.finish(to_json(response))

    @tornado.web.authenticated
    async def patch(self):
        self.set_header("Content-Type", "application/json")

        is_enabled = self.get_argument("enabled").lower()
        domain_name = self.get_argument("domain")
        user_access = await self.get_user_access_credentials()
        async with project_microservice_service.access_credentials(user_access):
            async with project_microservice_service.incoming_request(self.request):
                if is_enabled == "true":
                    service_response = await project_microservice_service.start_app(domain_name, user_access.username)
                elif is_enabled == "false":
                    service_response = await project_microservice_service.stop_app(domain_name, user_access.username)
                else:
                    service_response = ServiceResponse(
                        content=None,
                        meta=MetaResponse(RequestError(True, "The enable parameter must either true or false string")),
                    )

        return self.finish(to_json(service_response))

    @tornado.web.authenticated
    async def put(self):
        self.set_header("Content-Type", "application/json")
        domain_name = self.get_argument("domain")
        user_access = await self.get_user_access_credentials()
        async with project_microservice_service.access_credentials(user_access):
            async with project_microservice_service.incoming_request(self.request):
                response = await project_microservice_service.restart_app(domain_name, user_access.username)
        if response.has_failed():
            self.set_status(500)
        return self.finish(to_json(response))


class ProjectActionHandler(PABaseHandler):
    @tornado.web.authenticated
    async def post(self, project: str, action: str):
        if not project:
            return self.finish(
                to_json(
                    ServiceResponse(
                        content=None,
                        meta=MetaResponse(RequestError(True, "Missing project_id and/or version_id")),
                    )
                ),
                status_code=httpx.codes.BAD_REQUEST,
            )

        self.set_header("Content-Type", "application/json")
        self.set_status(200)
        api_version = self.request.headers.get("X-MS-API-Version", MS_API_Version.V1)

        user_access = await self.get_user_access_credentials()
        async with project_microservice_service.access_credentials(user_access):
            async with project_microservice_service.incoming_request(self.request):
                if action == "publish":
                    response = await project_microservice_service.publish_latest_version(project)
                    return self.finish(to_json(response))
                elif action == "delete":
                    self.set_status(200)
                    response = await project_microservice_service.delete_project(project, api_version)
                    if response.has_failed():
                        self.set_status(400)
                    return self.finish(to_json(response))
                else:
                    response = {"success": False, "message": f"Unknown action {action}"}
                    self.set_status(400)

        self.finish(json.dumps(response))

    @tornado.web.authenticated
    async def get(self, domain_name: str, log_type: str):
        """
        Fetch Panel app custom_logger.
        Parameters
        ----------
        domain_name: app domain
        log_type: [access_log | error_log]

        Returns
        -------
        JSON response
        """
        self.set_header("Content-Type", "application/json")

        response = await panel_service.get_app_log(self.api, domain_name, log_type)

        return self.finish(to_json(response))


class ProjectVersionHandler(NotebookBaseHandler):
    @tornado.web.authenticated
    async def get(self, project_id: UUID, version_id: UUID):
        if not project_id or not version_id:
            return self.finish(
                to_json(
                    ServiceResponse(
                        content=None,
                        meta=MetaResponse(RequestError(True, "Missing project_id and/or version_id")),
                    )
                ),
                status_code=httpx.codes.BAD_REQUEST,
            )

        self.set_header("Content-Type", "application/json")
        api_version = self.request.headers.get("X-MS-API-Version", MS_API_Version.V1)

        user_access = await self.get_user_access_credentials()
        async with project_microservice_service.access_credentials(user_access):
            async with project_microservice_service.incoming_request(self.request):
                response = await project_microservice_service.get_project_notebook(project_id, version_id, api_version)

        self.finish(to_json(response))

    @tornado.web.authenticated
    async def post(self, project_id: UUID, version_id: UUID):
        if version_id:
            self.finish(
                to_json(
                    ServiceResponse(
                        content=None,
                        meta=MetaResponse(RequestError(True, "POST method not supported")),
                    )
                ),
                status_code=httpx.codes.METHOD_NOT_ALLOWED,
            )

        if not project_id:
            self.finish(
                to_json(
                    ServiceResponse(
                        content=None,
                        meta=MetaResponse(RequestError(True, "Missing project_id")),
                    )
                ),
                status_code=httpx.codes.BAD_REQUEST,
            )

        self.set_header("Content-Type", "application/json")

        title = self.get_argument("title")
        description = self.get_argument("description")
        path = self.get_argument("path")
        version = self.get_argument("version")

        notebook_path = self._get_path(path)
        notebook_file_info = await self._get_file_info(notebook_path)
        user_access = await self.get_user_access_credentials()

        # validate project version
        if not project_service.is_valid_version(version):
            return self.finish(
                to_json(
                    ServiceResponse(
                        content=None,
                        meta=MetaResponse(RequestError(True, "Project version is not valid")),
                    )
                ),
                status_code=httpx.codes.BAD_REQUEST,
            )

        # validate notebook file path
        if not notebook_file_info:
            return self.finish(
                to_json(
                    ServiceResponse(
                        content=None,
                        meta=MetaResponse(RequestError(True, "Project file path not found")),
                    )
                ),
                status_code=httpx.codes.BAD_REQUEST,
            )

        # validate onda environment
        try:
            environment = await self.extract_notebook_environment(notebook_path)
        except CondaEnvironmentNotFound:
            return self.finish(
                to_json(
                    ServiceResponse(
                        content=None,
                        meta=MetaResponse(RequestError(True, "Conda environment not found")),
                    )
                ),
                status_code=httpx.codes.BAD_REQUEST,
            )

        err = RequestError(status=False)

        custom_logger.info("Creating project version info")
        project_info = await project_service.create_project_info(
            title,
            version,
            description,
            notebook_file_info,
            environment,
        )

        custom_logger.info("Generating project files")
        project_content = await project_service.generate_files(
            user_access.username,
            title,
            version,
            project_info,
            environment,
            notebook_path,
        )
        try:
            api_version = self.request.headers.get("X-MS-API-Version", MS_API_Version.V1)
            async with project_microservice_service.access_credentials(user_access):
                async with project_microservice_service.incoming_request(self.request):
                    custom_logger.info("Creating a new project version")
                    request_data = CreateProjectVersion(
                        project_id=project_id,
                        version=version,
                        environment=environment,
                        notebook_file_name=notebook_file_info.file_name,
                        notebook_file_size=notebook_file_info.file_size,
                    )

                    project_version_response = await project_microservice_service.create_project_version(
                        request_data, project_content.archive_content, api_version
                    )
        except CantCreateProjectVersion:
            project_version_response = "Can't create new project version"
            err.status = True
            err.message = "Failure while creating new project version"

        response = to_json(ServiceResponse(meta=MetaResponse(error=err), content=project_version_response))
        self.finish(response)

    @tornado.web.authenticated
    async def delete(self, project_id: UUID, version_id: UUID):
        if not project_id or not version_id:
            return self.finish(
                to_json(
                    ServiceResponse(
                        content=None,
                        meta=MetaResponse(RequestError(True, "Missing project_id and/or version_id")),
                    )
                ),
                status_code=httpx.codes.BAD_REQUEST,
            )

        self.set_header("Content-Type", "application/json")
        api_version = self.request.headers.get("X-MS-API-Version", MS_API_Version.V1)

        user_access = await self.get_user_access_credentials()
        async with project_microservice_service.access_credentials(user_access):
            async with project_microservice_service.incoming_request(self.request):
                response = await project_microservice_service.delete_project_version(
                    project_id, version_id, api_version
                )

        self.finish(to_json(response))


class FeaturedAppsHandler(BackendHandler):
    @tornado.web.authenticated
    async def get(self):
        try:
            response = await backend_proxy(
                self.request,
                f"{config['cloud']['staticContent']}/anaconda-notebooks/panels/featured-apps/featured-apps.json",
            )
            self.finish(response)
        except Exception:
            custom_logger.error("Not possible to get featured apps")


def get_routes(base_url: str) -> _RuleList:
    handlers: Dict[str, Type[PABaseHandler]] = {
        "healthz": HealthzRouteHandler,
        "deployments": DeploymentHandler,
        "project-info": ProjectInfoHandler,
        "projects": ProjectHandler,
        "projects/download": ProjectDownloadHandler,
        r"projects/(.*?)/version/(.*?)": ProjectVersionHandler,
        r"projects/(.*?)/(.*)/": ProjectActionHandler,
        "featured-apps": FeaturedAppsHandler,
    }
    rules_new_prefix = create_rules(base_url, "aext_panels_server", handlers)
    rules_old_prefix = create_rules(base_url, "pythonanywhere", handlers)

    return rules_new_prefix + rules_old_prefix



import json
from dataclasses import asdict, dataclass, field
from io import BytesIO
from typing import Any, List, Optional
from uuid import UUID

# @TODO verify if we still need this one and remove


@dataclass
class GenerateFiles:
    project_info_content: BytesIO
    archive_content: BytesIO


@dataclass
class PAProjectInfo:
    title: str
    project_id: str
    notebook_file_name: str
    notebook_file_size: str
    environment: str
    path: str
    description: str = ""
    last_published_at: str = ""
    versions: List[str] = field(default_factory=list)

    def to_bytes_io(self):
        return BytesIO(json.dumps(asdict(self)).encode("utf-8"))


@dataclass
class AppVersion:
    id: UUID
    version: str
    created_at: str
    updated_at: str
    last_published_at: Optional[str] = None

    @classmethod
    def create_from_microservice_response(self, json_response: dict) -> "ProjectInfoData":
        return AppVersion(
            id=json_response["id"],
            version=json_response["version"],
            created_at=json_response["created_at"],
            updated_at=json_response["updated_at"],
            last_published_at=json_response["last_published_at"],
        )


@dataclass
class VersioningData:
    next_version: str
    current_version: str = ""
    all_versions: List[AppVersion] = field(default_factory=list)


@dataclass
class Domain:
    webapp_id: int
    domain: str
    user: str
    enabled: bool


@dataclass
class ProjectInfoData:
    title: str
    project_id: str
    versioning: VersioningData
    notebook_file_name: str
    notebook_file_size: str
    permission: str
    environment: str
    status: str = ""
    description: str = ""
    last_published_at: str = ""
    domains: List[Domain] = field(default_factory=list)

    def is_active(self):
        # TODO : THIS SHOULD BECOME A ENUM
        return self.status == "published"

    @classmethod
    def create_from_microservice_response(self, json_response: dict) -> "ProjectInfoData":
        return ProjectInfoData(
            title=json_response["title"],
            project_id=json_response["id"],
            versioning=VersioningData(
                next_version=json_response["versioning"]["next_version"],
                current_version=json_response["versioning"]["current_version"],
                all_versions=json_response["versioning"]["all_versions"],
            ),
            notebook_file_name=json_response["notebook_file_name"],
            notebook_file_size=json_response["notebook_file_size"],
            permission=json_response["permission"],
            environment=json_response["environment"],
            description=json_response["description"],
            last_published_at=json_response["last_published_at"],
            domains=json_response["domains"],
            status=json_response["status"],
        )


@dataclass
class FileInfo:
    file_name: str
    file_size: str


@dataclass
class RequestError:
    status: bool
    message: str = ""


@dataclass
class Deployment:
    id: int = None
    user: str = None
    domains: List[Domain] = field(default_factory=list)
    command: Optional[str] = None
    domain_name: Optional[str] = None
    webapp: Optional[int] = None
    enabled: Optional[bool] = None


@dataclass
class ListDeploymentsResponse:
    deployments: List[Deployment] = field(default_factory=list)


@dataclass
class MetaResponse:
    error: RequestError


@dataclass
class ServiceResponse:
    meta: MetaResponse
    content: Any

    def has_succeeded(self) -> bool:
        return not self.meta.error.status

    def has_failed(self) -> bool:
        return self.meta.error.status

    def set_error_message(self, msg: str) -> bool:
        if not self.meta and not self.meta.error:
            return False

        self.meta.error.message = msg
        return True

    def get_error_message(self) -> str:
        if not self.meta and not self.meta.error:
            return ""

        return self.meta.error.message

    def set_error_status(self, status: bool) -> bool:
        if not self.meta and not self.meta.error:
            return False
        self.meta.error.status = status
        return True


#########################################################################
#                       REQUEST SCHEMAS                                 #
#########################################################################
@dataclass
class Webapp:
    command: str


@dataclass
class CreatePanelAppRequest:
    domain_name: str
    enabled: bool
    webapp: Webapp


@dataclass
class PublishAppRequest:  # let's redeclare it just in case PA change the contract
    enabled: bool = True


@dataclass
class UnpublishAppRequest:  # let's redeclare it just in case PA change the contract
    enabled: bool = False


#########################################################################
#                       RESPONSE SCHEMAS                                #
#########################################################################
@dataclass
class WebappResponse:
    id: int
    domain_name: str
    enabled: bool
    status: Optional[str] = None


@dataclass
class CreatePanelAppResponse:
    app: WebappResponse


@dataclass
class ListAppsResponse:
    num_active_apps: int
    num_max_apps: int
    apps: List[ProjectInfoData]



import base64
import contextlib
from dataclasses import fields
from typing import IO, Iterator

import httpx
from aext_panels_server.api import (
    ProjectDeploymentMicroserviceAPI,
    project_deployment_ms_api,
)
from aext_panels_server.api_v2 import project_deployment_ms_api_v2
from aext_panels_server.const import MS_API_Version
from aext_panels_server.exceptions import (
    CantCreateProject,
    CantCreateProjectVersion,
    CantUpdateProject,
)
from aext_panels_server.handlers_helpers import create_response
from aext_panels_server.logger import custom_logger
from aext_panels_server.request_schemas.notebook_projects_microservice import (
    CreateProject,
    CreateProjectVersion,
    UpdateProject,
)
from aext_panels_server.schemas import (
    AppVersion,
    CreatePanelAppResponse,
    Deployment,
    Domain,
    ListAppsResponse,
    ProjectInfoData,
    ServiceResponse,
    WebappResponse,
)
from tornado.httputil import HTTPServerRequest

from aext_shared.handler import UserAccessCredentials


class ProjectDeploymentMicroserviceService:
    @contextlib.asynccontextmanager
    async def access_credentials(self, user_credentials: UserAccessCredentials) -> None:
        custom_logger.info(f"Set user creds: {user_credentials}")
        ProjectDeploymentMicroserviceStrategy.set_access_credentials(user_credentials)
        yield
        for field in fields(user_credentials):
            setattr(user_credentials, field.name, None)
        ProjectDeploymentMicroserviceStrategy.set_access_credentials(user_credentials)

    @contextlib.asynccontextmanager
    async def incoming_request(self, incoming_request: HTTPServerRequest) -> None:
        ProjectDeploymentMicroserviceStrategy.set_incoming_request(incoming_request)
        yield
        ProjectDeploymentMicroserviceStrategy.set_incoming_request(None)

    async def create_project(
        self, request_data: CreateProject, archive_content: IO, api_version: MS_API_Version
    ) -> ProjectInfoData:
        """
        Creates a new project
        Parameters
        ----------
        request_data: data to be sent to the microservice
        archive_content: archive content

        Returns ProjectInfoData
        -------
        """
        project_deployment_ms_api = ProjectDeploymentMicroserviceStrategy.get_api(api_version)
        response = await project_deployment_ms_api.create_project(request_data, archive_content)
        if response.status_code == httpx.codes.CREATED:
            return ProjectInfoData.create_from_microservice_response(response.json())

        raise CantCreateProject

    async def upload_project_archive(self, project_id: str, archive_content: IO) -> bool:
        """
        Uploads an archive to the latest version of the project
        Parameters
        ----------
        project_id: id of the project
        archive_content: file content

        Returns boolean
        -------

        """
        response = await project_deployment_ms_api.upload_project_archive(project_id, archive_content)
        if response.status_code == httpx.codes.OK:
            return True
        return False

    async def update_project(
        self, project_id: str, request_data: UpdateProject, archive_content: IO
    ) -> ProjectInfoData:
        """
        Update project information
        Parameters
        ----------
        project_id: project id that will be updated
        request_data: data  used to update the project that should be sent to the microservice
        archive_content: zip file containing the new version

        Returns ProjectInfoData
        -------

        """
        if not await self.upload_project_archive(project_id, archive_content):
            custom_logger.error("Could not update project: archive upload failed")
            raise CantUpdateProject

        response = await project_deployment_ms_api.update_project(project_id, request_data)
        if response.status_code == httpx.codes.OK:
            return ProjectInfoData.create_from_microservice_response(response.json())

        custom_logger.error(f"Could not update project: {response}")
        raise CantUpdateProject

    async def download_project_files(
        self, project_id: str, project_version: str, api_version: MS_API_Version
    ) -> ServiceResponse:
        """
        Downloads project archive files
        Parameters
        ----------
        project_id: id of the project
        project_version: version of the project

        Returns service response containing the archive content
        -------

        """
        project_deployment_ms_api = ProjectDeploymentMicroserviceStrategy.get_api(api_version)
        response = await project_deployment_ms_api.download_files(project_id, project_version)
        if response.status_code == httpx.codes.OK:
            if response.headers["content-type"] != "application/json":
                json_response = base64.b64encode(response.content).decode("utf-8")
            else:
                json_response = response.json()

            return create_response(
                status=False,
                content={"archive": json_response["archive"]},
                json_response=False,
            )

        return create_response(status=True, content={}, json_response=False)

    async def list_projects(self, user_id: str) -> ServiceResponse:
        """
        List all user apps

        Parameters
        ----------
        user_id: id of the user

        Returns
        -------

        """
        custom_logger.debug("Listing user projects")
        response = await project_deployment_ms_api.list_projects(user_id)
        custom_logger.debug(f"MICROSERVICE Response: {response}")
        if not response:
            return create_response(status=True, content={}, json_response=False)

        if response.status_code == httpx.codes.OK:
            custom_logger.debug("Listing user projects")
            json_response = response.json()
            all_apps = [
                ProjectInfoData.create_from_microservice_response(project_data)
                for project_data in json_response["projects"]
            ]
            return create_response(
                status=False,
                content=ListAppsResponse(
                    num_active_apps=json_response["num_active_apps"],
                    num_max_apps=json_response["num_max_apps"],
                    apps=all_apps,
                ),
                json_response=False,
            )

        return create_response(status=True, content={}, json_response=False)

    def _handle_app_action_response(self, json_response: dict, user_id: str) -> ServiceResponse:
        """
        This function handle the response returned by the microservice for
        start_app, stop_app and restart_app
        """
        domains = [
            Domain(
                webapp_id=json_response["webapp_id"],
                domain=domain["name"],
                enabled=domain["enabled"],
                user=user_id,
            )
            for domain in json_response["domains"]
        ]
        deployment = Deployment(
            id=json_response["id"],
            domains=domains,
            webapp=json_response["webapp_id"],
            enabled=True,
        )
        return create_response(
            status=False,
            content=deployment,
            json_response=False,
        )

    async def start_app(self, app_domain_name: str, user_id: str) -> ServiceResponse:
        """
        Start a deployed app that was previously stopped
        Parameters
        ----------
        app_domain_name: domain associated with the app
        user_id: id of the user

        Returns ServiceResponse
        -------

        """
        response = await project_deployment_ms_api.start_app(app_domain_name)
        if response.status_code == httpx.codes.OK:
            return self._handle_app_action_response(response.json(), user_id)

        return create_response(
            status=True,
            content={},
            json_response=False,
        )

    async def stop_app(self, app_domain_name: str, user_id: str) -> ServiceResponse:
        """
        Stops a deployed app

        Parameters
        ----------
        app_domain_name: domain associated with the app
        user_id: the app owner id

        Returns
        -------

        """
        response = await project_deployment_ms_api.stop_app(app_domain_name)
        if response.status_code == httpx.codes.OK:
            return self._handle_app_action_response(response.json(), user_id)

        return create_response(
            status=True,
            content={},
            json_response=False,
        )

    async def restart_app(self, app_domain_name: str, user_id: str) -> ServiceResponse:
        """
        Restart a deployed app
        Parameters
        ----------
        app_domain_name: domain associated with the app
        user_id: the app owner id

        Returns
        -------

        """
        response = await project_deployment_ms_api.restart_app(app_domain_name)
        if response.status_code == httpx.codes.OK:
            return self._handle_app_action_response(response.json(), user_id)

        return create_response(
            status=True,
            content={},
            json_response=False,
        )

    async def publish_latest_version(self, project_id: str) -> ServiceResponse:
        """
        Publishes/deploy the latest project version.

        Parameters
        ----------
        project_id: id of the project

        Returns ServiceResponse
        -------

        """
        api_response = await project_deployment_ms_api.publish_latest_version(project_id)
        if api_response.status_code in (httpx.codes.OK, httpx.codes.CREATED):
            json_response = api_response.json()
            if json_response["domains"]:
                response_model = CreatePanelAppResponse(
                    app=WebappResponse(
                        id=json_response["domains"][0]["webapp"],
                        domain_name=json_response["domains"][0]["domain"],
                        enabled=json_response["domains"][0]["enabled"],
                    )
                )
                return create_response(
                    status=False,
                    content=response_model,
                    json_response=False,
                )

        return create_response(
            status=True,
            content={},
            json_response=False,
        )

    async def delete_project(self, project_id: str, api_version: MS_API_Version) -> ServiceResponse:
        """
        Delete project files and possible deployments if there are any

        Parameters
        ----------
        project_id: id of the project

        Returns ServiceResponse
        -------

        """
        project_deployment_ms_api = ProjectDeploymentMicroserviceStrategy.get_api(api_version)
        response = await project_deployment_ms_api.delete_project(project_id)
        if response.status_code == httpx.codes.NO_CONTENT:
            return create_response(status=False, content={}, json_response=False)

        return create_response(status=True, content={}, json_response=False)

    async def create_project_version(
        self, request_data: CreateProjectVersion, archive_content: IO, api_version: MS_API_Version
    ) -> ServiceResponse:
        """
        Creates a new project version

        Parameters
        ----------
        request_data: data to be sent to the microservice
        archive_content: archive content

        Returns AppVersion
        -------
        """
        project_deployment_ms_api = ProjectDeploymentMicroserviceStrategy.get_api(api_version)
        response = await project_deployment_ms_api.create_project_version(request_data, archive_content)
        if response.status_code == httpx.codes.CREATED:
            return AppVersion.create_from_microservice_response(response.json())

        raise CantCreateProjectVersion

    async def get_project_notebook(self, project_id: str, version_id: str, api_version: MS_API_Version):
        """
        Retrieve a project notebook

        Parameters
        ----------
        project_id: id of the project
        version_id: version of the notebook to be retrieved

        Returns ServiceResponse
        ----------
        """
        project_deployment_ms_api = ProjectDeploymentMicroserviceStrategy.get_api(api_version)
        response = await project_deployment_ms_api.get_project_notebook(project_id, version_id)
        if response.status_code == httpx.codes.OK:
            return create_response(status=False, content=response.json(), json_response=False)

        return create_response(status=True, content={}, json_response=False)

    async def delete_project_version(self, project_id: str, version_id: str, api_version: MS_API_Version):
        """
        Delete a project version

        Parameters
        ----------
        project_id: id of the project
        version_id: version to be deleted

        Returns ServiceResponse
        ----------
        """
        project_deployment_ms_api = ProjectDeploymentMicroserviceStrategy.get_api(api_version)
        response = await project_deployment_ms_api.delete_project_version(project_id, version_id)
        if response.status_code == httpx.codes.NO_CONTENT:
            return create_response(status=False, content=response.json(), json_response=False)

        return create_response(status=True, content={}, json_response=False)


class ProjectDeploymentMicroserviceStrategy:
    API = {
        "v1": project_deployment_ms_api,
        "v2": project_deployment_ms_api_v2,
    }

    @classmethod
    def get_api(cls, version: MS_API_Version) -> ProjectDeploymentMicroserviceAPI:
        try:
            return cls.API.get(version, MS_API_Version.V1)
        except IndexError as e:
            msg = "Unknown API version"
            raise ValueError(msg) from e

    @classmethod
    def apis(cls) -> Iterator[ProjectDeploymentMicroserviceAPI]:
        for api in cls.API.values():
            yield api

    @classmethod
    def set_access_credentials(cls, user_access_credentials: UserAccessCredentials):
        for api in cls.apis():
            api.set_anaconda_cloud_access_token(user_access_credentials.access_token)
            api.set_anaconda_cloud_username(user_access_credentials.username)

    @classmethod
    def set_incoming_request(cls, incoming_request: HTTPServerRequest):
        for api in cls.apis():
            api.set_incoming_request(incoming_request)


project_microservice_service = ProjectDeploymentMicroserviceService()



import logging

from traitlets.config import Application


class CustomLogger:
    jlab_logger = Application.instance().log
    general_logger = logging.getLogger(__name__)

    def _log(self, log_type: str, msg: str):
        getattr(self.jlab_logger, f"{log_type}")(msg)
        getattr(self.general_logger, f"{log_type}")(msg)

    def _unset_loggers(self):
        self.jlab_logger = None
        self.general_logger = None

    def info(self, msg: str):
        self._log("info", msg)

    def warning(self, msg: str):
        self._log("warning", msg)

    def error(self, msg: str):
        self._log("error", msg)

    def debug(self, msg: str):
        self._log("debug", msg)


custom_logger = CustomLogger()



import httpx
from aext_panels_server.api import PythonAnywhereAPI
from aext_panels_server.const import ACCESS_LOG, ERROR_LOG, PROJECT_LOG_PATH, SERVER_LOG
from aext_panels_server.schemas import MetaResponse, RequestError, ServiceResponse
from aext_panels_server.utils import call_api


class PanelService:
    async def get_app_log(self, api: PythonAnywhereAPI, domain: str, log_type: str) -> ServiceResponse:
        """
        Panel app custom_logger. This function will fetch log file for a Panel app.
        Parameters
        ----------
        api: PythonAnywhereAPI
        domain: Project domain
        log_type: [access_log|error_log]

        Returns
        -------
        ServiceResponse
        """
        if log_type not in [ACCESS_LOG, ERROR_LOG, SERVER_LOG]:
            return ServiceResponse(
                meta=MetaResponse(
                    error=RequestError(
                        status=True,
                        message="Project log does not exit",
                    )
                ),
                content=None,
            )

        log_file_name = {
            ACCESS_LOG: f"{domain}.access.log",
            SERVER_LOG: f"{domain}.server.log",
            ERROR_LOG: f"{domain}.error.log",
        }

        log_path = f"{PROJECT_LOG_PATH}/{log_file_name[log_type]}"

        err, response = await call_api(
            api.download_file(log_path, include_root_path=False),
            always_return_response=True,
        )

        if not err.status:
            return ServiceResponse(
                meta=MetaResponse(error=RequestError(status=False)),
                content=response.text,
            )

        if not response:
            return ServiceResponse(
                meta=MetaResponse(
                    error=RequestError(
                        status=True,
                        message="Unexpected error while fetching project log",
                    )
                ),
                content=None,
            )

        if err.status:
            err_resp = ServiceResponse(
                meta=MetaResponse(
                    error=RequestError(
                        status=True,
                    )
                ),
                content=None,
            )
            if response.status_code == httpx.codes.NOT_FOUND:
                err_resp.set_error_message("Index file not found")
            elif response.status_code == httpx.codes.UNAUTHORIZED:
                err_resp.set_error_message("Failed to fetch index - Unauthorized")
            else:
                err_resp.set_error_message(f"Failed to fetch index - {err.message}")
            return err_resp
        return ServiceResponse(meta=MetaResponse(error=RequestError(status=False)), content=response.text)


panel_service = PanelService()



"""
List extracted from: https://en.wikipedia.org/wiki/List_of_snakes_by_common_name
"""

SNAKES = [
    "Viper Adder",
    "Common adder",
    "Death Adder",
    "Desert death adder",
    "Horned adder",
    "Long-nosed adder",
    "Many-horned adder",
    "Mountain adder",
    "Mud adder",
    "Namaqua dwarf adder",
    "Peringuey's adder",
    "Puff adder",
    "African puff adder",
    "Rhombic night adder",
    "Sand adder",
    "Dwarf sand adder",
    "Namib dwarf sand adder",
    "Water adder",
    "Aesculapian snake",
    "Anaconda",
    "Bolivian anaconda",
    "De Schauensee's anaconda",
    "Green anaconda",
    "Yellow anaconda",
    "Arafura file snake",
    "Asp",
    "European asp",
    "Egyptian asp",
    "African beaked snake",
    "Ball Python",
    "Bird snake",
    "Black-headed snake",
    "Mexican black kingsnake",
    "Black rat snake",
    "Black snake",
    "Red-bellied black snake",
    "Blind snake",
    "Brahminy blind snake",
    "Texas blind snake",
    "Western blind snake",
    "Boa",
    "Abaco Island boa",
    "Amazon tree boa",
    "Boa constrictor",
    "Cuban boa",
    "Dumeril's boa",
    "Dwarf boa",
    "Emerald tree boa",
    "Hogg Island boa",
    "Jamaican boa",
    "Madagascar ground boa",
    "Madagascar tree boa",
    "Puerto Rican boa",
    "Rainbow boa",
    "Red-tailed boa",
    "Rosy boa",
    "Rubber boa",
    "Sand boa",
    "Tree boa",
    "Boiga",
    "Boomslang",
    "Brown snake",
    "Eastern brown snake",
    "Bull snake",
    "Bushmaster",
    "Dwarf beaked snake",
    "Rufous beaked snake",
    "Canebrake",
    "Cantil",
    "Cascabel",
    "Cat-eyed snake",
    "Banded cat-eyed snake",
    "Green cat-eyed snake",
    "Cat snake",
    "Andaman cat snake",
    "Beddome's cat snake",
    "Dog-toothed cat snake",
    "Forsten's cat snake",
    "Gold-ringed cat snake",
    "Gray cat snake",
    "Many-spotted cat snake",
    "Nicobar cat snake",
    "Sri Lanka cat snake",
    "Tawny cat snake",
    "Chicken snake",
    "Coachwhip snake",
    "Cobra",
    "Andaman cobra",
    "Arabian cobra",
    "Asian cobra",
    "Banded water cobra",
    "Black-necked cobra",
    "Black-necked spitting cobra",
    "Black tree cobra",
    "Burrowing cobra",
    "Cape cobra",
    "Caspian cobra",
    "Chinese cobra",
    "Cobra de capello",
    "Congo water cobra",
    "Common cobra",
    "Eastern water cobra",
    "Egyptian cobra",
    "Equatorial spitting cobra",
    "False cobra",
    "False water cobra",
    "Forest cobra",
    "Gold tree cobra",
    "Indian cobra",
    "Indochinese spitting cobra",
    "Javan spitting cobra",
    "King cobra",
    "Mandalay cobra",
    "Monocled cobra",
    "Monoculate cobra",
    "Mozambique spitting cobra",
    "North Philippine cobra",
    "Nubian spitting cobra",
    "Philippine cobra",
    "Red spitting cobra",
    "Rinkhals cobra",
    "Shield-nosed cobra",
    "Sinai desert cobra",
    "Southern Indonesian spitting cobra",
    "Southern Philippine cobra",
    "Southwestern black spitting cobra",
    "Snouted cobra",
    "Spectacled cobra",
    "Spitting cobra",
    "Storm water cobra",
    "Thai cobra",
    "Taiwan cobra",
    "West African brown spitting cobra",
    "White-lipped cobra",
    "Yellow cobra",
    "Zebra spitting cobra",
    "Collett's snake",
    "Congo snake",
    "Copperhead",
    "American copperhead",
    "Australian copperhead",
    "Coral snake",
    "Arizona coral snake",
    "Beddome's coral snake",
    "Brazilian coral snake",
    "Cape coral snake",
    "Eastern coral snake",
    "False coral snake[disambiguation needed]",
    "Harlequin coral snake",
    "High Woods coral snake",
    "Malayan long-glanded coral snake",
    "Texas Coral Snake",
    "Western coral snake",
    "Corn snake",
    "South eastern corn snake",
    "Cottonmouth",
    "Crowned snake",
    "Cuban wood snake",
    "Eastern hognose snake",
    "Egg-eater",
    "Indian egg-eater",
    "Eyelash viper",
    "Eastern coral snake",
    "Fer-de-lance",
    "Fierce snake",
    "Fishing snake",
    "Flying snake",
    "Golden tree snake",
    "Indian flying snake",
    "Moluccan flying snake",
    "Ornate flying snake",
    "Paradise flying snake",
    "Twin-Barred tree snake",
    "Banded Flying Snake",
    "Fox snake, three species of Pantherophis",
    "Forest flame snake",
    "Garter snake",
    "Checkered garter snake",
    "Common garter snake",
    "San Francisco garter snake",
    "Texas garter snake",
    "Glossy snake",
    "Gopher snake",
    "Cape gopher snake",
    "Grass snake",
    "Green snake",
    "Rough green snake",
    "Smooth green snake",
    "Ground snake",
    "Common ground snake",
    "Three-lined ground snake",
    "Western ground snake",
    "Habu",
    "Himehabu",
    "Okinawan habu",
    "Sakishima habu",
    "Tokara habu",
    "Harlequin snake",
    "Elaps harlequin snake",
    "Herald snake",
    "Hognose snake",
    "Blonde hognose snake",
    "Dusty hognose snake",
    "Eastern hognose snake",
    "Jan's hognose snake",
    "Giant Malagasy hognose snake",
    "Mexican hognose snake",
    "Plains hognose snake",
    "Ringed hognose snake",
    "South American hognose snake",
    "Southern hognose snake",
    "Speckled hognose snake",
    "Tri-color hognose snake",
    "Western hognose snake",
    "Hoop snake",
    "Hundred pacer",
    "Ikaheka snake",
    "Indigo snake",
    "Jamaican Tree Snake",
    "Jararacussu",
    "Keelback",
    "Andrea's keelback",
    "Asian keelback",
    "Assam keelback",
    "Black-striped keelback",
    "Buff striped keelback",
    "Burmese keelback",
    "Checkered keelback",
    "Common keelback",
    "Hill keelback",
    "Himalayan keelback",
    "Khasi Hills keelback",
    "Modest keelback",
    "Nicobar Island keelback",
    "Nilgiri keelback",
    "Orange-collared keelback",
    "Red-necked keelback",
    "Sikkim keelback",
    "Speckle-bellied keelback",
    "Tiger keelback",
    "Wall's keelback",
    "White-lipped keelback",
    "Wynaad keelback",
    "Yunnan keelback",
    "King brown",
    "King cobra",
    "King snake",
    "California kingsnake",
    "Desert kingsnake",
    "Grey-banded kingsnake",
    "North eastern king snake",
    "Prairie kingsnake",
    "Scarlet kingsnake",
    "Speckled kingsnake",
    "Krait",
    "Banded krait",
    "Blue krait",
    "Black krait",
    "Burmese krait",
    "Ceylon krait",
    "Indian krait",
    "Lesser black krait[1]",
    "Malayan krait",
    "Many-banded krait",
    "Northeastern hill krait[2]",
    "Red-headed krait",
    "Sind krait",
    "South Andaman krait",
    "Suzhen's krait",
    "Large shield snake",
    "Lancehead",
    "Common lancehead",
    "Lora",
    "Grey Lora",
    "Lyre snake",
    "Baja California lyresnake",
    "Central American lyre snake",
    "Texas lyre snake",
    "Eastern lyre snake",
    "Machete savane",
    "Mamba",
    "Black mamba",
    "Green mamba",
    "Eastern green mamba",
    "Western green mamba",
    "Mamushi",
    "Mangrove snake",
    "Milk snake",
    "Moccasin snake",
    "Montpellier snake",
    "Mud snake",
    "Eastern mud snake",
    "Western mud snake",
    "Mussurana",
    "Night snake",
    "Cat-eyed night snake",
    "Texas night snake",
    "Nichell snake",
    "Narrowhead Garter Snake",
    "Nose-horned viper",
    "Rhinoceros viper",
    "Vipera ammodytes",
    "Parrot snake",
    "Mexican parrot snake",
    "Patchnose snake",
    "Perrotet's shieldtail snake",
    "Pine snake",
    "Pipe snake",
    "Asian pipe snake",
    "Dwarf pipe snake",
    "Red-tailed pipe snake",
    "Python",
    "African rock python",
    "Amethystine python",
    "Angolan python",
    "Australian scrub python",
    "Ball python",
    "Bismarck ringed python",
    "Black headed python",
    "Blood python",
    "Boelen python",
    "Borneo short-tailed python",
    "Bredl's python",
    "Brown water python",
    "Burmese python",
    "Calabar python",
    "Western carpet python",
    "Centralian carpet python",
    "Coastal carpet python",
    "Inland carpet python",
    "Jungle carpet python",
    "New Guinea carpet python",
    "Northwestern carpet python",
    "Southwestern carpet python",
    "Children's python",
    "Dauan Island water python",
    "Desert woma python",
    "Diamond python",
    "Flinders python",
    "Green tree python",
    "Halmahera python",
    "Indian python",
    "Indonesian water python",
    "Macklot's python",
    "Moluccan python",
    "Oenpelli python",
    "Olive python",
    "Papuan python",
    "Pygmy python",
    "Red blood python",
    "Reticulated python",
    "Kayaudi dwarf reticulated python",
    "Selayer reticulated python",
    "Rough-scaled python",
    "Royal python",
    "Savu python",
    "Spotted python",
    "Stimson's python",
    "Sumatran short-tailed python",
    "Tanimbar python",
    "Timor python",
    "Wetar Island python",
    "White-lipped python",
    "Brown white-lipped python",
    "Northern white-lipped python",
    "Southern white-lipped python",
    "Woma python",
    "Western woma python",
    "Queen snake",
    "Racer",
    "Bimini racer",
    "Buttermilk racer",
    "Eastern racer",
    "Eastern yellowbelly sad racer",
    "Mexican racer",
    "Southern black racer",
    "Tan racer",
    "West Indian racer",
    "Raddysnake",
    "Southwestern blackhead snake",
    "Rat snake",
    "Baird's rat snake",
    "Beauty rat snake",
    "Great Plains rat snake",
    "Green rat snake",
    "Japanese forest rat snake",
    "Japanese rat snake",
    "King rat snake",
    "Mandarin rat snake",
    "Persian rat snake",
    "Red-backed rat snake[3]",
    "Twin-spotted rat snake",
    "Yellow-striped rat snake",
    "Manchurian Black Water Snake",
    "Rattlesnake",
    "Arizona black rattlesnake",
    "Aruba rattlesnake",
    "Chihuahuan ridge-nosed rattlesnake",
    "Coronado Island rattlesnake",
    "Durango rock rattlesnake",
    "Dusky pigmy rattlesnake",
    "Eastern diamondback rattlesnake",
    "Grand Canyon rattlesnake",
    "Great Basin rattlesnake",
    "Hopi rattlesnake",
    "Lance-headed rattlesnake",
    "Long-tailed rattlesnake",
    "Massasauga rattlesnake",
    "Mexican green rattlesnake",
    "Mexican west coast rattlesnake",
    "Midget faded rattlesnake",
    "Mojave rattlesnake",
    "Northern black-tailed rattlesnake",
    "Oaxacan small-headed rattlesnake",
    "Rattler",
    "Red diamond rattlesnake",
    "Southern Pacific rattlesnake",
    "Southwestern speckled rattlesnake",
    "Tancitaran dusky rattlesnake",
    "Tiger rattlesnake",
    "Timber rattlesnake",
    "Tropical rattlesnake",
    "Twin-spotted rattlesnake",
    "Uracoan rattlesnake",
    "Western diamondback rattlesnake",
    "Ribbon snake",
    "Rinkhals",
    "River jack",
    "Sea snake",
    "Annulated sea snake",
    "Beaked sea snake",
    "Dubois's sea snake",
    "Hardwicke's sea snake",
    "Hook Nosed Sea Snake",
    "Olive sea snake",
    "Pelagic sea snake",
    "Stoke's sea snake",
    "Yellow-banded sea snake",
    "Yellow-bellied sea snake",
    "Yellow-lipped sea snake",
    "Shield-tailed snake",
    "Sidewinder",
    "Colorado desert sidewinder",
    "Mojave desert sidewinder",
    "Sonoran sidewinder",
    "Small-eyed snake",
    "Smooth snake",
    "Brazilian smooth snake",
    "European smooth snake",
    "Stiletto snake",
    "Striped snake",
    "Japanese striped snake",
    "Sunbeam snake",
    "Taipan",
    "Central ranges taipan",
    "Coastal taipan",
    "Inland taipan",
    "Paupan taipan",
    "Tentacled snake",
    "Tic polonga",
    "Tiger snake",
    "Chappell Island tiger snake",
    "Common tiger snake",
    "Down's tiger snake",
    "Eastern tiger snake",
    "King Island tiger snake",
    "Krefft's tiger snake",
    "Peninsula tiger snake",
    "Tasmanian tiger snake",
    "Western tiger snake",
    "Tigre snake",
    "Tree snake",
    "Blanding's tree snake",
    "Blunt-headed tree snake",
    "Brown tree snake",
    "Long-nosed tree snake",
    "Many-banded tree snake",
    "Northern tree snake",
    "Trinket snake",
    "Black-banded trinket snake",
    "Twig snake",
    "African twig snake",
    "Twin Headed King Snake",
    "Titanoboa",
    "Urutu",
    "Vine snake",
    "Asian Vine Snake, Whip Snake",
    "American Vine Snake",
    "Mexican vine snake",
    "Viper",
    "Asp viper",
    "Bamboo viper",
    "Bluntnose viper",
    "Brazilian mud Viper",
    "Burrowing viper",
    "Bush viper",
    "Great Lakes bush viper",
    "Hairy bush viper",
    "Nitsche's bush viper",
    "Rough-scaled bush viper",
    "Spiny bush viper",
    "Carpet viper",
    "Crossed viper",
    "Cyclades blunt-nosed viper",
    "Eyelash viper",
    "False horned viper",
    "Fea's viper",
    "Fifty pacer",
    "Gaboon viper",
    "Hognosed viper",
    "Horned desert viper",
    "Horned viper",
    "Jumping viper",
    "Kaznakov's viper",
    "Leaf-nosed viper",
    "Leaf viper",
    "Levant viper",
    "Long-nosed viper",
    "McMahon's viper",
    "Mole viper",
    "Nose-horned viper",
    "Rhinoceros viper",
    "Vipera ammodytes",
    "Palestine viper",
    "Pallas' viper",
    "Palm viper",
    "Amazonian palm viper",
    "Black-speckled palm-pitviper",
    "Eyelash palm-pitviper",
    "Green palm viper",
    "Mexican palm-pitviper",
    "Guatemalan palm viper",
    "Honduran palm viper",
    "Siamese palm viper",
    "Side-striped palm-pitviper",
    "Yellow-lined palm viper",
    "Pit viper",
    "Banded pitviper",
    "Bamboo pitviper",
    "Barbour's pit viper",
    "Black-tailed horned pit viper",
    "Bornean pitviper",
    "Brongersma's pitviper",
    "Brown spotted pitviper[4]",
    "Cantor's pitviper",
    "Elegant pitviper",
    "Eyelash pit viper",
    "Fan-Si-Pan horned pitviper",
    "Flat-nosed pitviper",
    "Godman's pit viper",
    "Green tree pit viper",
    "Habu pit viper",
    "Hagen's pitviper",
    "Horseshoe pitviper",
    "Jerdon's pitviper",
    "Kanburian pit viper",
    "Kaulback's lance-headed pitviper",
    "Kham Plateau pitviper",
    "Large-eyed pitviper",
    "Malabar rock pitviper",
    "Malayan pit viper",
    "Mangrove pit viper",
    "Mangshan pitviper",
    "Motuo bamboo pitviper",
    "Nicobar bamboo pitviper",
    "Philippine pitviper",
    "Pointed-scaled pit viper[5]",
    "Red-tailed bamboo pitviper",
    "Schultze's pitviper",
    "Stejneger's bamboo pitviper",
    "Sri Lankan pit viper",
    "Temple pit viper",
    "Tibetan bamboo pitviper",
    "Tiger pit viper",
    "Undulated pit viper",
    "Wagler's pit viper",
    "Wirot's pit viper",
    "Portuguese viper",
    "Rhinoceros viper",
    "River jack",
    "Russell's viper",
    "Sand viper",
    "Saw-scaled viper",
    "Schlegel's viper",
    "Sedge viper",
    "Sharp-nosed viper",
    "Snorkel viper",
    "Temple viper",
    "Tree viper",
    "Chinese tree viper",
    "Guatemalan tree viper",
    "Hutton's tree viper",
    "Indian tree viper",
    "Large-scaled tree viper",
    "Malcolm's tree viper",
    "Nitsche's tree viper",
    "Pope's tree viper",
    "Rough-scaled tree viper",
    "Rungwe tree viper",
    "Sumatran tree viper",
    "White-lipped tree viper",
    "Ursini's viper",
    "Western hog-nosed viper",
    "Wart snake",
    "Water moccasin",
    "Water snake",
    "Bocourt's water snake",
    "Northern water snake",
    "Whip snake",
    "Long-nosed whip snake",
    "Wolf snake",
    "African wolf snake",
    "Barred wolf snake",
    "Worm snake",
    "Common worm snake",
    "Longnosed worm snake",
    "Wutu",
    "Yarara",
    "Zebra snake",
]



"""
List of adjectives
"""

ADJECTIVES = [
    "Accommodating",
    "Adventurous",
    "Affectionate",
    "Agreeable",
    "Alluring",
    "Altruistic",
    "Ambitious",
    "Amorous",
    "Amusing",
    "Appealing",
    "Ardent",
    "Attentive",
    "Attractive",
    "Audacious",
    "Authentic",
    "Awesome",
    "Balanced",
    "Benevolent",
    "Bold",
    "Brave",
    "Bright",
    "Capable",
    "Captivating",
    "Caring",
    "Charitable",
    "Charming",
    "Charming",
    "Cheerful",
    "Cheery",
    "Clever",
    "Comforting",
    "Comical",
    "Committed",
    "Compassionate",
    "Concerned",
    "Considerate",
    "Consoling",
    "Content",
    "Cooperative",
    "Courageous",
    "Courteous",
    "Creative",
    "Dashing",
    "Decent",
    "Delightful",
    "Dependable",
    "Devoted",
    "Diligent",
    "Dutiful",
    "Eager",
    "Echarming",
    "Empathetic",
    "Enchanting",
    "Endearing",
    "Energetic",
    "Engrossing",
    "Enjoyable",
    "Enjoyable",
    "Enterprising",
    "Enthusiastic",
    "Ethical",
    "Excited",
    "Extravagantly",
    "Faithful",
    "Fanciful",
    "Fantastic",
    "Fearless",
    "Fervent",
    "Funny",
    "Generous",
    "Gentle",
    "Genuine",
    "Giving",
    "Good-natured",
    "Happy",
    "Hard-working",
    "Helpful",
    "Heroic",
    "Honest",
    "Honorable",
    "Hospitable",
    "Humane",
    "Humorous",
    "Imaginary",
    "Imaginative",
    "Impassioned",
    "Industrious",
    "Ingenious",
    "Innovative",
    "Inspired",
    "Inspiring",
    "Intelligent",
    "Intrepid",
    "Inventive",
    "Joyous",
    "Keen",
    "Lavish",
    "Level-headed",
    "Liberal",
    "Likable",
    "Lively",
    "Lovely",
    "Loving",
    "Loyal",
    "Magnanimous",
    "Mindful",
    "Motivated",
    "Munificent",
    "Nice",
    "Novel",
    "Obliging",
    "Original",
    "Passionate",
    "Perceptive",
    "Persistent",
    "Personable",
    "Pioneering",
    "Pleasant",
    "Polite",
    "Popular",
    "Practical",
    "Pragmatic",
    "Profound",
    "Quick",
    "Quick-witted",
    "Rational",
    "Real",
    "Realistic",
    "Reliable",
    "Resourceful",
    "Respectful",
    "Responsible",
    "Riveting",
    "Secure",
    "Seductive",
    "Sensational",
    "Sensible",
    "Sharp",
    "Sincere",
    "Smart",
    "Solid",
    "Spirited",
    "Steadfast",
    "Steady",
    "Supportive",
    "Sympathetic",
    "Tenacious",
    "Tender",
    "Thorough",
    "Thoughtful",
    "True",
    "Trustworthy",
    "Truthful",
    "Unafraid",
    "Untroubled",
    "Upright",
    "Valiant",
    "Venturous",
    "Vigorous",
    "Virtuous",
    "Vivacious",
    "Well-behaved",
    "Wholehearted",
    "Witty",
]



from aext_panels_server.services.project_index import project_index_service


async def test_get_project_info_file():
    project_index_path = project_index_service.get_project_index_file_path()
    assert project_index_path == ".projects/index.json"

    project_index_path = project_index_service.get_project_index_file_path(include_root_path=True)

    assert project_index_path == "/var/www/panels/.projects/index.json"



import zipfile
from io import BytesIO
from unittest import mock

import pytest
import yaml
from aext_panels_server.api import project_deployment_ms_api
from aext_panels_server.exceptions import NbconvertNotInstalled, PanelNotInstalled
from aext_panels_server.schemas import FileInfo, PAProjectInfo
from aext_panels_server.services.project import project_service
from aext_panels_server.utils import clean_notebook, get_relative_file_path

from .fixtures import get_file_full_path


@pytest.fixture
def mocked_conda_list():
    return [
        {
            "base_url": "https://conda.anaconda.org/pypi",
            "build_number": 0,
            "build_string": "pypi_0",
            "channel": "pypi",
            "dist_name": "panel-1.0.4-pypi_0",
            "name": "panel",
            "platform": "pypi",
            "version": "1.0.4",
        },
        {
            "base_url": "https://conda.anaconda.org/pypi",
            "build_number": 0,
            "build_string": "pypi_0",
            "channel": "pypi",
            "dist_name": "nbconvert-0.0.4-pypi_0",
            "name": "nbconvert",
            "platform": "pypi",
            "version": "0.0.4",
        },
    ]


@pytest.fixture
def mock_conda_list(mocked_conda_list):
    with mock.patch("aext_panels_server.api.CondaAPI.list") as m:
        m.return_value = mocked_conda_list
        yield m


@pytest.mark.parametrize(
    "_status_code, _json_resp",
    [
        (
            200,
            {
                "versions": ["2023.1.1", "2023.1.2"],
                "description": "dummy",
                "title": "dummy-name",
                "path": "/home/user",
            },
        )
    ],
)
async def test_create_project_info_should_return_success(
    mock_python_anywhere_api,
    python_anywhere_api,
    _json_resp,
):
    notebook_file_info = FileInfo(
        file_name=get_file_full_path("assets/test-notebook.ipynb"),
        file_size="926 Bytes",
    )

    project_info = await project_service.create_project_info(
        title="dummy-project",
        version="2023.3.1",
        description="My new description",  # Modifying the description
        notebook_file_info=notebook_file_info,
        environment="base",
    )

    assert project_info.description != _json_resp["description"]
    assert len(project_info.versions) < len(_json_resp["versions"])


@pytest.mark.parametrize(
    "_status_code, _json_resp",
    [
        (
            200,
            {
                "title": "dummy-project",
                "project_id": "ad80abf2-9767-45bf-8c29-ec9699f2474e",
                "notebook_file_name": get_relative_file_path(get_file_full_path("assets/test-notebook.ipynb")),
                "notebook_file_size": "926 Bytes",
                "environment": "base",
                "path": ".projects/ad80abf2-9767-45bf-8c29-ec9699f2474e/project_info.json",
                "description": "My new description",
                "last_published_at": "",
                "versions": ["2023.3.1"],
            },
        )
    ],
)
async def test_get_project_info_should_return_successful(mock_python_anywhere_api, python_anywhere_api):
    notebook_file_info = FileInfo(
        file_name=get_file_full_path("assets/test-notebook.ipynb"),
        file_size="926 Bytes",
    )

    project_info = await project_service.create_project_info(
        title="dummy-project",
        version="2023.3.1",
        description="My new description",  # Modifying the description
        notebook_file_info=notebook_file_info,
        environment="base",
    )

    response = await project_service.get_project_info(
        api=python_anywhere_api,
        project_id=project_info.project_id,
    )
    assert project_info.title == response.title
    assert project_info.versions == response.versions
    assert project_info.description == response.description
    assert project_info.notebook_file_name == response.notebook_file_name
    assert project_info.notebook_file_size == response.notebook_file_size
    assert project_info.environment == response.environment


async def test_generate_zip_file(
    mock_python_anywhere_api,
    python_anywhere_api,
    _json_resp,
    mock_conda_list,
):
    project_name = "dummy-project"
    version = "2022.04.01"
    notebook = "test-notebook.ipynb"
    nb_path = get_file_full_path(f"assets/{notebook}")
    project_zip = await project_service.generate_zip_file("anaconda-user", project_name, version, "base", nb_path)
    assert isinstance(project_zip, BytesIO)
    files = {}
    with zipfile.ZipFile(project_zip) as zip_obj:
        for zipinfo in zip_obj.infolist():
            with zip_obj.open(zipinfo) as f:
                files[zipinfo.filename] = f.read()
    proj_yaml_file = f"{project_name}-{version}/conda-project.yaml"
    assert proj_yaml_file in files
    proj_yaml = yaml.load(files[proj_yaml_file], yaml.SafeLoader)
    assert proj_yaml == {
        "name": project_name,
        "maintainers": ["anaconda-user"],
        "platforms": ["linux-64"],
        "environments": {"default": ["environment.yaml"]},
        "commands": {"serve": {"cmd": f"panel serve {notebook}"}},
    }
    env_file = f"{project_name}-{version}/environment.yaml"
    assert env_file in files
    nb_file = f"{project_name}-{version}/{notebook}"
    assert nb_file in files
    with open(nb_path, "r") as f:
        assert files[nb_file].decode("utf-8") == clean_notebook(f.read())


@pytest.mark.parametrize(
    "mocked_conda_list",
    [
        [
            {
                "base_url": "https://conda.anaconda.org/pypi",
                "build_number": 0,
                "build_string": "pypi_0",
                "channel": "pypi",
                "dist_name": "notpanel-1.0.4-pypi_0",
                "name": "notpanel",
                "platform": "pypi",
                "version": "1.0.4",
            }
        ]
    ],
)
async def test_generate_zip_file_should_contain_panels(
    mock_python_anywhere_api,
    python_anywhere_api,
    _json_resp,
    mock_conda_list,
):
    project_name = "dummy-project"
    version = "2022.04.01"
    notebook = "test-notebook.ipynb"
    nb_path = get_file_full_path(f"assets/{notebook}")
    with pytest.raises(PanelNotInstalled):
        await project_service.generate_zip_file("anaconda-user", project_name, version, "base", nb_path)


@pytest.mark.parametrize(
    "mocked_conda_list",
    [
        [
            {
                "base_url": "https://conda.anaconda.org/pypi",
                "build_number": 0,
                "build_string": "pypi_0",
                "channel": "pypi",
                "dist_name": "panel-1.0.4-pypi_0",
                "name": "panel",
                "platform": "pypi",
                "version": "1.0.4",
            }
        ]
    ],
)
async def test_generate_zip_file_should_have_nbconvert(
    mock_python_anywhere_api,
    python_anywhere_api,
    _json_resp,
    mock_conda_list,
    mocked_conda_list,
):
    project_name = "dummy-project"
    version = "2022.04.01"
    notebook = "test-notebook.ipynb"
    nb_path = get_file_full_path(f"assets/{notebook}")
    with pytest.raises(NbconvertNotInstalled):
        await project_service.generate_zip_file("anaconda-user", project_name, version, "base", nb_path)


async def test_generate_files(project_info, mock_conda_list):
    project_info["notebook_file_name"] = get_file_full_path("assets/test-notebook.ipynb")
    project_info["notebook_file_size"] = "926 Bytes"

    project_content = await project_service.generate_files(
        "anaconda-user",
        project_info["title"],
        project_info["versions"][-1],
        PAProjectInfo(**project_info),
        project_info["environment"],
        get_file_full_path("assets/test-notebook.ipynb"),
    )

    assert isinstance(project_content.project_info_content, BytesIO)
    assert isinstance(project_content.archive_content, BytesIO)


async def test_project_versions():
    valid_versions = ["2023.04.1", "2023.04.12", "2023.05.31", "2023.05.212"]
    invalid_versions = ["2023.04.01", "2023.04", "2023.05.031", ""]

    for version in valid_versions:
        assert project_service.is_valid_version(version) is True

    for version in invalid_versions:
        assert project_service.is_valid_version(version) is False


def test_set_anaconda_cloud_access_token():
    project_deployment_ms_api.set_anaconda_cloud_access_token("user-access-token")
    assert project_deployment_ms_api.anaconda_cloud_access_token == "user-access-token"



from __future__ import annotations

import json
import logging
import os
from dataclasses import asdict, dataclass
from subprocess import run
from typing import IO, Dict, Optional

import httpx
from aext_panels_server.const import (
    DEFAULT_HTTPX_TIMEOUT_SEC,
    DEV_USE_STAGING_MICROSERVICE,
    PROJECT_ROOT_PATH,
    PYTHONANYWHERE_PRODUCTION_DOMAIN,
    PYTHONANYWHERE_PRODUCTION_SITE,
    Environment,
)
from aext_panels_server.exceptions import CantRetrieveCondaPackages
from aext_panels_server.logger import custom_logger
from aext_panels_server.request_schemas.notebook_projects_microservice import (
    CreateProject,
    CreateProjectVersion,
    UpdateProject,
)
from aext_panels_server.schemas import (
    CreatePanelAppRequest,
    PublishAppRequest,
    UnpublishAppRequest,
)
from aext_panels_server.utils import get_environment, handle_network_exceptions
from anaconda_cloud_auth.client import AuthConfig
from anaconda_cloud_auth.token import TokenInfo
from httpx import Timeout
from tornado.httputil import HTTPServerRequest
from traitlets.config import Application

from aext_shared.backend_proxy import backend_proxy_core
from aext_shared.errors import BackendError

logger = logging.getLogger(__name__)


@dataclass
class HeadersAndCookies:
    headers: Optional[Dict] = None
    cookies: Optional[Dict] = None


class CondaAPI:
    @staticmethod
    def list(environment):
        if environment == os.environ.get("CONDA_DEFAULT_ENV"):
            command = ["conda", "list", "--json"]
        else:
            command = ["conda", "list", "--json", "--name", environment]
        proc = run(command, text=True, capture_output=True)
        result = json.loads(proc.stdout)

        if "error" in result:
            raise CantRetrieveCondaPackages(f"Failed retrieving Conda Packages. {result['error']}")

        return result


class BaseAPI:
    incoming_request = None

    @property
    def pa_domain(self) -> str:
        return os.getenv("PYTHONANYWHERE_DOMAIN", PYTHONANYWHERE_PRODUCTION_DOMAIN)

    @property
    def pa_site(self) -> str:
        return os.getenv("PYTHONANYWHERE_SITE", PYTHONANYWHERE_PRODUCTION_SITE)

    def _get_headers_and_cookies(self, **kwargs) -> HeadersAndCookies:
        raise NotImplementedError()

    def _get_url(self, endpoint: str, *args, **kwargs) -> str:
        raise NotImplementedError()

    async def _simple_request(
        self,
        method,
        endpoint,
        params=None,
        data=None,
        json_data=None,
        timeout_sec=None,
        **kwargs,
    ):
        request_kwargs = {}
        headers_and_cookies = self._get_headers_and_cookies(**kwargs)
        url = self._get_url(endpoint, **kwargs)
        if timeout_sec is None:
            timeout_sec = DEFAULT_HTTPX_TIMEOUT_SEC
        try:
            custom_logger.info(
                f"--> <REQUEST {method.upper()} {url} params:{params} data:{data} json-data:{json_data}> timeout: {timeout_sec} headers:{headers_and_cookies.headers} cookies:{headers_and_cookies.cookies}"
            )

            if params:
                request_kwargs["params"] = params
            if headers_and_cookies.headers:
                request_kwargs["headers"] = headers_and_cookies.headers
            if headers_and_cookies.cookies:
                request_kwargs["cookies"] = headers_and_cookies.cookies

            if method.lower() in ["post", "put", "patch"]:
                # httpx does not take data argument for all methods. Adding for the ones relevant
                if data:
                    request_kwargs["data"] = data
                if kwargs.get("files"):
                    request_kwargs["files"] = kwargs.get("files")
                if json_data:
                    request_kwargs["json"] = json_data

            if self.incoming_request is not None:
                custom_logger.info("Using backend proxy")
                custom_logger.info(f"--> Request KWARGS: {request_kwargs}")
                request_kwargs["method"] = method.upper()
                httpx_response = (await backend_proxy_core(self.incoming_request, url, **request_kwargs))[1]

            else:
                # TODO: remove this once we are done with tests and safe with backend-proxy
                custom_logger.info("Using async client")
                async with httpx.AsyncClient(
                    headers=headers_and_cookies.headers,
                    cookies=headers_and_cookies.cookies,
                    timeout=Timeout(timeout_sec),
                    follow_redirects=True,
                ) as client:
                    httpx_response = await getattr(client, method)(url, **request_kwargs)
                    res_content = httpx_response.text
                    custom_logger.info(
                        f"--> <RESPONSE status:{httpx_response.status_code} reason:{httpx_response.reason_phrase} content:{res_content}>"
                    )
            return httpx_response
        except BackendError as ex:
            custom_logger.error(f"--> Proxy request error: {vars(ex)}")
        except Exception as ex:
            custom_logger.error(f"--> Error while making simple request: {method} {url}")
            custom_logger.debug(f"--> Exception: {ex}")
            raise

    async def _get(self, endpoint, params=None, timeout_sec=None, **kwargs):
        return await self._simple_request("get", endpoint, params=params, timeout_sec=timeout_sec, **kwargs)

    async def _put(self, endpoint, data=None, json_data=None, timeout_sec=None, **kwargs):
        return await self._simple_request(
            "put", endpoint, data=data, json_data=json_data, timeout_sec=timeout_sec, **kwargs
        )

    async def _patch(self, endpoint, params=None, json_data=None, timeout_sec=None, **kwargs):
        return await self._simple_request(
            "patch",
            endpoint,
            params=params,
            json_data=json_data,
            timeout_sec=timeout_sec,
            **kwargs,
        )

    async def _post(self, endpoint, data=None, json_data=None, timeout_sec=None, **kwargs):
        return await self._simple_request(
            "post", endpoint, data=data, json_data=json_data, timeout_sec=timeout_sec, **kwargs
        )

    async def _delete(self, endpoint: str, data: Dict = None, timeout_sec=None, **kwargs):
        return await self._simple_request("delete", endpoint, data=data, timeout_sec=timeout_sec, **kwargs)


class PythonAnywhereAPI(BaseAPI):
    """
    Implements an API to make requests to PythonAnywhere APIs.

    See details at https://help.pythonanywhere.com/pages/API/

    Attributes
    ----------
    domain: str | None
        The PythonAnywhere domain to make requests on.
    username: str | None
        The username making the request.
    token: str | None
        An API token authorizing the user. See
        https://www.pythonanywhere.com/user/<username>/account/#api_token
    csrftoken: str | None
        CSRF Token to authenticate request.
    sessionidstr: str | None
        Session ID cookie used to identify authorized requests.
    """

    _endpoints = {
        "webapps": "/api/v1/user/{username}/webapps/",
        "webapp": "/api/v1/user/{username}/webapps/{app_id}",
        "domains": "/api/v1/user/{username}/domains/{domain_name}/",
        "list_files": "/api/v0/user/{username}/files/tree/?path={path}",
        "file_access": "/api/v0/user/{username}/files/path{path}",
        "delete_file": "/api/v0/user/{username}/files/path{path}",
        "panels": "/api/v1/user/{username}/panels/",
        "panel": "/api/v1/user/{username}/panels/{domain_name}/",
        "max_apps": "/api/v0/user/{username}/features/",
    }

    _root_path = PROJECT_ROOT_PATH

    def __init__(
        self,
        domain: str | None = None,
        username: str | None = None,
        token: str | None = None,
        csrftoken: str | None = None,
        sessionid: str | None = None,
    ):
        if not domain:
            domain = os.getenv("PYTHONANYWHERE_SITE", PYTHONANYWHERE_PRODUCTION_SITE)
        self.domain = domain

        if not username:
            username = os.getenv("PYTHONANYWHERE_USER")

        if not token:
            token = os.getenv("PYTHONANYWHERE_TOKEN", "")

        self.username = username
        self.token = token
        self.csrftoken = csrftoken
        self.sessionid = sessionid
        self.session = None
        self.log = Application.instance().log

    def _get_headers_and_cookies(self, **kwargs) -> HeadersAndCookies:
        """
        Returns a pair of headers and cookies used to create the request session
        """
        headers = {
            "Referer": f"https://{self.domain}/user/{self.username}",
        }
        if self.token:
            headers["Authorization"] = f"Token {self.token}"
        else:
            headers["X-CSRFToken"] = self.csrftoken

        cookies = {}
        if self.sessionid:
            cookies["sessionid"] = self.sessionid
        if self.csrftoken:
            cookies["csrftoken"] = self.csrftoken

        custom_logger.info("Creating request session")
        return HeadersAndCookies(headers=headers, cookies=cookies)

    def _get_url(self, endpoint: str, **kwargs) -> str:
        endpoint = endpoint.format(username=self.username, **kwargs)
        return f"https://{self.domain}{endpoint}"

    # Public API

    async def list_deployments(self):
        """
        Returns the list of deployments

        Returns
        -------
        deployments: List[Dict[str, Any]]

        Raises
        ------
        HTTPError
        """
        return await self._get(self._endpoints["webapps"])

    async def get_deployment_info(self, app_id: int) -> httpx.Response:
        """
        Return information about a web app's configuration

        Returns
        -------
        deployments: List[Dict[str, Any]]

        Raises
        ------
        HTTPError
        """
        custom_logger.info(f"Getting app {app_id} info")
        return await self._get(self._endpoints["webapp"], app_id=app_id)

    async def get_max_apps_info(self, user_id: str) -> httpx.Response:
        """
        Return information about how many apps a user can deploy

        Returns
        -------
        features: Dict[str, Any]

        Raises
        ------
        HTTPError
        """
        custom_logger.info(f"Getting max apps info for user {user_id}")
        return await self._get(self._endpoints["max_apps"], user_id=user_id)

    async def delete_deployment(self, domain_name: str) -> httpx.Response:
        """
        Deletes a deployment (a deployed app which may be inactive or active)
        and all domains associated with it
        Parameters
        ----------
        domain_name: deployed app domain name

        Returns httpx response object
        -------

        """
        custom_logger.info(f"Deleting app deployment {domain_name}")
        return await self._delete(self._endpoints["panel"], domain_name=domain_name)

    async def delete_project(self, project_id: str) -> httpx.Response:
        """
        Deletes all files related to the project/app that are stored in PA file storage
        Parameters
        ----------
        project_id: project identifier

        Returns httpx response object
        -------

        """
        custom_logger.info(f"Deleting app project {project_id}")
        return await self._delete(
            self._endpoints["delete_file"],
            path=f"{self._root_path}.projects/{project_id}",
        )

    async def add_deployment(self, domain_name, **params):
        """
        Return information about a web app's configuration

        Parameters
        ----------
        domain_name: str
            Name of the domain to host the deployment at.
        python_version: str
            Python version in the base environment.
        source_directory: str
            Directory where the application source lives.
        virtualenv_path: str
            Virtual environment to activate
        force_https: bool
            Whether to force an HTTPS redirect
        password_protection_enabled: bool
            Whether to password protection to the app.
        password_protection_username: str
            Username to authorize with the password protected application.
        password_protection_password: str
            Password to authorize with the password protected application.

        Raises
        ------
        HTTPError
        """
        return await self._put(self._endpoints["app"], domain_name=domain_name)

    async def list_files(self, path):
        """
        List files at the given path.

        Parameters
        ----------
        path: str
          The path for which to list files.

        Returns
        -------
        files: List[str]

        Raises
        ------
        HTTPError
        """
        return await self._get(self._endpoints["list_files"], path=f"{self._root_path}/{path}")

    async def download_file(self, path, include_root_path: bool = True):
        """
        List files at the given path.

        Parameters
        ----------
        path: str
          The path for which to list files.

        include_root_path: bool
          The root path will be appended

        Returns
        -------
        files: List[str]

        Raises
        ------
        HTTPError
        """
        base_path = ""
        if include_root_path:
            base_path = self._root_path

        return await self._get(self._endpoints["file_access"], path=f"{base_path}{path}")

    async def unpublish_app(self, request_data: UnpublishAppRequest, domain_name: str) -> httpx.Response:
        """
        Given a published app this function unpublish it by flipping the `enabled` flag for
        a given UnpublishAppRequest (domain, enabled, command).

        Returns

        Parameters
        ----------
        request_data
        domain_name

        Returns httpx.Response
        -------

        """
        custom_logger.info(f"Unpublishing panel app {request_data}")
        return await self._patch(self._endpoints["panel"], json_data=asdict(request_data), domain_name=domain_name)

    async def publish_app(self, request_data: PublishAppRequest, domain_name: str) -> httpx.Response:
        """
        Given an unpublished app this function publishes it by flipping the `enabled` flag for
        a given PublishAppRequest (domain, enabled, command).

        Parameters
        ----------
        request_data
        domain_name

        Returns httpx.Response
        -------

        """
        custom_logger.info(f"Publishing panel App {request_data}")
        return await self._patch(self._endpoints["panel"], json_data=asdict(request_data), domain_name=domain_name)

    async def upload_file(self, path, fileobj: IO) -> httpx.Response:
        """
        Uploads a file object to a given path.

        Warning: If the file in the given exists it will be overwritten.

        Parameters
        ----------
        path: str
            Path relative to upload the file to.
        fileobj: IO
            File object containing the object to be uploaded.

        Returns
        -------
        response
        """
        custom_logger.info(f"Uploading file {path}")
        files = {"content": ("content", fileobj)}
        return await self._post(self._endpoints["file_access"], path=path, files=files)

    async def create_panel_app(self, request_data: CreatePanelAppRequest) -> httpx.Response:
        """
        Creates a new panel app given a CreatePanelAppRequest (domain, enabled, command)
        Parameters
        ----------
        request_data

        Returns
        -------
        """
        custom_logger.info(f"Creating panel app {request_data}")
        return await self._post(self._endpoints["panels"], json_data=asdict(request_data))


class ProjectDeploymentMicroserviceAPI(BaseAPI):
    NUCLEUS_BASE_URLS = {
        Environment.PRODUCTION: "https://anaconda.cloud/api/panels/v1/",
        Environment.STAGING: "https://nucleus-latest.anacondaconnect.com/api/panels/v1/",
        Environment.LOCAL: os.getenv("DEV_PANELS_MICROSERVICE_ADDRESS", "http://notebook-deployment-api:8000/v1/"),
    }

    anaconda_cloud_access_token = None
    anaconda_cloud_username = None
    incoming_request = None
    pa_username = None

    def __init__(self, environment: Environment) -> None:
        try:
            self.base_url = self.NUCLEUS_BASE_URLS[environment]
        except KeyError:
            logger.error(f"Could not find neither an API key nor Nucleus URL for {environment}")
            raise ValueError(f"Could not find neither an API key nor Nucleus URL for {environment}")

    def set_anaconda_cloud_access_token(self, access_token: Optional[str]) -> None:
        self.anaconda_cloud_access_token = access_token

    def set_anaconda_cloud_username(self, username: Optional[str]) -> None:
        self.anaconda_cloud_username = username

    def set_incoming_request(self, incoming_request: Optional[HTTPServerRequest]) -> None:
        self.incoming_request = incoming_request

    def _get_url(self, endpoint, **kwargs) -> str:
        """
        Get the URL used in the request
        Parameters
        ----------
        endpoint: the API endpoint
        kwargs

        Returns full url
        -------

        """
        return f"{self.base_url}{endpoint}"

    def _get_headers_and_cookies(self, **kwargs) -> HeadersAndCookies:
        """
        Get request headers
        Returns: Dict containing all necessary headers
        -------

        """
        pa_username = os.getenv("JUPYTERHUB_USER", os.getenv("PYTHONANYWHERE_USER", self.anaconda_cloud_username))
        pa_api_token = os.getenv("API_TOKEN", os.getenv("PYTHONANYWHERE_TOKEN"))
        cloud_flare_client = os.getenv("NUCLEUS_CLOUDFLARE_CLIENT_ID")
        cloud_flare_secret = os.getenv("NUCLEUS_CLOUDFLARE_CLIENT_SECRET")
        pa_site = self.pa_site
        pa_domain = self.pa_domain

        if pa_username is None:
            raise ValueError("PA Username not set")

        headers = {
            "pa-username": pa_username,
            "pa-site": pa_site,
            "pa-domain": pa_domain,
        }

        if pa_api_token:
            headers["pa-api-token"] = pa_api_token

        # Authorization header
        # Currently works with both Panel MS API KEY and anaconda.cloud access token
        assert self.anaconda_cloud_access_token, "Anaconda cloud access token missing"
        headers["Authorization"] = f"Bearer {self.anaconda_cloud_access_token}"
        custom_logger.info(f"Authorization: Anaconda Cloud Access Token {headers}")

        if cloud_flare_client and cloud_flare_secret:
            headers.update(
                {
                    "CF-Access-Client-Id": cloud_flare_client,
                    "CF-Access-Client-Secret": cloud_flare_secret,
                }
            )

        return HeadersAndCookies(
            headers=headers,
            cookies={},
        )

    async def get_project(self, project_id: str, **kwargs) -> httpx.Response:
        endpoint = f"project/{project_id}/"
        return await self._get(endpoint, **kwargs)

    @handle_network_exceptions(max_retries=1)
    async def create_project(self, request_data: CreateProject, archive_content: IO, **kwargs) -> httpx.Response:
        """
        Creates a new project
        Parameters
        ----------
        request_data: all the data necessary to create the project
        archive_content: archive content (bytes)

        Returns http response
        -------

        """
        endpoint = "project/"
        archive = {"archive": archive_content}
        return await self._post(endpoint, data=asdict(request_data), files=archive, **kwargs)

    @handle_network_exceptions(max_retries=1)
    async def upload_project_archive(self, project_id: str, archive_content: IO, **kwargs) -> httpx.Response:
        """
        Uploads project's archive

        Parameters
        ----------
        project_id: id of the project
        archive_content: archive content (bytes)

        Returns http response
        -------

        """
        endpoint = f"project/{project_id}/archive/upload/"
        archive = {"archive": archive_content}
        return await self._post(endpoint, files=archive, **kwargs)

    @handle_network_exceptions(max_retries=1)
    async def update_project(self, project_id: str, request_data: UpdateProject, **kwargs) -> httpx.Response:
        """
        Updates the project information
        Parameters
        ----------
        project_id: project to be updated
        request_data: the new project data
        kwargs

        Returns http response
        -------

        """
        endpoint = f"project/{project_id}/"
        return await self._patch(endpoint, json_data={"project_data": asdict(request_data)})

    @handle_network_exceptions(max_retries=2, default_timeout=4)
    async def list_projects(self, user_id: str, **kwargs) -> httpx.Response:
        """
        List all users project
        Parameters
        ----------
        user_id: id of the user
        kwargs

        Returns http repsonse
        -------

        """
        endpoint = f"user/{user_id}/projects/"
        return await self._get(endpoint, **kwargs)

    @handle_network_exceptions(max_retries=1)
    async def delete_project(self, project_id: str, **kwargs) -> httpx.Response:
        """
        Deletes a project
        Parameters
        ----------
        project_id: project id that should be deleted
        kwargs

        Returns http response
        -------

        """
        endpoint = f"project/{project_id}/"
        return await self._delete(endpoint, **kwargs)

    @handle_network_exceptions(max_retries=1)
    async def publish_latest_version(self, project_id: str, **kwargs) -> httpx.Response:
        """
        Publish the latest project saved version
        Parameters
        ----------
        project_id: project that should have its latest version published
        kwargs

        Returns http response
        -------

        """
        endpoint = f"project/{project_id}/publish/"
        return await self._post(endpoint, data={"domain": self.pa_domain}, **kwargs)

    @handle_network_exceptions(max_retries=1)
    async def stop_app(self, app_domain_name: str, **kwargs) -> httpx.Response:
        """
        Given an up and running app this function stops the app.
        Parameters
        ----------
        app_domain_name: domain name associated to the app that should be stopped
        kwargs

        Returns http response
        -------

        """
        endpoint = f"deployment/{app_domain_name}/unpublish/"
        return await self._post(endpoint, **kwargs)

    @handle_network_exceptions(max_retries=1)
    async def start_app(self, app_domain_name: str, **kwargs) -> httpx.Response:
        """
        This function start a stopped app
        Parameters
        ----------
        app_domain_name: domain name associated to the app that should be stopped
        kwargs

        Returns http response
        -------

        """
        endpoint = f"deployment/{app_domain_name}/publish/"
        return await self._post(endpoint, **kwargs)

    @handle_network_exceptions(max_retries=1)
    async def restart_app(self, app_domain_name: str, **kwargs) -> httpx.Response:
        """
        Given an up and running app this function restarts it by stopping and starting the app
        Parameters
        ----------
        app_domain_name: domain name associated to the app that should be stopped
        kwargs

        Returns http response
        -------

        """
        endpoint = f"deployment/{app_domain_name}/restart/"
        return await self._post(endpoint, **kwargs)

    @handle_network_exceptions(max_retries=1)
    async def download_files(self, project_id: str, project_version: str, **kwargs) -> httpx.Response:
        """
        Downloads the project files (archive)
        Parameters
        ----------
        project_id: project that should be downloaded
        project_version: the version of the project that should be downloaded
        kwargs

        Returns http response
        -------

        """
        endpoint = f"project/{project_id}/version/{project_version}/download/"
        return await self._get(endpoint, **kwargs)

    @handle_network_exceptions(max_retries=1)
    async def create_project_version(
        self, request_data: CreateProjectVersion, archive_content: IO, **kwargs
    ) -> httpx.Response:
        """
        Creates a new project version

        Parameters
        ----------
        project_id: project id that will have a new version created
        request_data: data necessary to create a new version
        kwargs

        Returns http response
        -------

        """
        endpoint = f"project/{request_data.project_id}/version/"
        archive = {"archive": archive_content}
        return await self._post(endpoint, data=asdict(request_data), files=archive, **kwargs)

    @handle_network_exceptions(max_retries=1)
    async def get_project_notebook(self, project_id: str, version_id: str, **kwargs) -> httpx.Response:
        raise NotImplementedError

    @handle_network_exceptions(max_retries=1)
    async def delete_project_version(self, project_id: str, version_id: str, **kwargs) -> httpx.Response:
        raise NotImplementedError


project_deployment_ms_api = ProjectDeploymentMicroserviceAPI(get_environment())



import json
import urllib
from dataclasses import asdict
from unittest import mock

import aext_panels_server as panel
import respx
import tornado.web
from aext_panels_server.api import HeadersAndCookies
from aext_panels_server.schemas import (
    FileInfo,
    MetaResponse,
    RequestError,
    ServiceResponse,
)
from jupyter_server.auth.identity import IdentityProvider
from tornado.testing import AsyncHTTPTestCase


class TestHealthzHandler(AsyncHTTPTestCase):
    def get_app(self):
        return tornado.web.Application(
            panel.get_routes("/"),
            cookie_secret="abc",
            identity_provider=IdentityProvider(token="abc"),
        )

    def test_healthz_handler(self):
        headers = {"Authorization": "token abc"}
        response = self.fetch("/aext_panels_server/healthz", headers=headers)
        assert response.code == 200
        expected = {"live": True}
        assert json.loads(response.body) == expected


class TestProjectDownloadHandler(AsyncHTTPTestCase):
    success = {"archive": "zipfiles"}
    error = "error"

    def get_app(self):
        return tornado.web.Application(
            panel.get_routes("/"),
            cookie_secret="abc",
            identity_provider=IdentityProvider(token="abc"),
        )

    @respx.mock
    def test_download_project_files_v2_success(self):
        respx.get().respond(status_code=200, json=self.success)
        headers = {"Authorization": "token abc", "Cookie": "access_token=xyz", "X-MS-API-Version": "v2"}

        with mock.patch("aext_panels_server.api.ProjectDeploymentMicroserviceAPI._get_headers_and_cookies") as m:
            m.return_value = HeadersAndCookies(headers=headers, cookies={})

            response = self.fetch(
                "/aext_panels_server/projects/download?project_id=test-project&project_version=test-version",
                headers=headers,
            )
            assert response.code == 200
            assert json.loads(response.body) == asdict(
                ServiceResponse(meta=MetaResponse(RequestError(False)), content=self.success)
            )


class TestProjectActionHandler(AsyncHTTPTestCase):
    success = "success"
    error = "error"

    def get_app(self):
        return tornado.web.Application(
            panel.get_routes("/"),
            cookie_secret="abc",
            identity_provider=IdentityProvider(token="abc"),
        )

    @respx.mock
    def test_delete_project_v2_success(self):
        respx.delete().respond(status_code=204, json=self.success)
        headers = {"Authorization": "token abc", "Cookie": "access_token=xyz", "X-MS-API-Version": "v2"}

        with mock.patch("aext_panels_server.api.ProjectDeploymentMicroserviceAPI._get_headers_and_cookies") as m:
            m.return_value = HeadersAndCookies(headers=headers, cookies={})

            response = self.fetch(
                "/aext_panels_server/projects/test-project/delete/", method="POST", headers=headers, body=""
            )
            assert response.code == 200
            assert json.loads(response.body) == asdict(
                ServiceResponse(meta=MetaResponse(RequestError(False)), content={})
            )


class TestProjectVersionHandler(AsyncHTTPTestCase):
    success = "success"
    error = "error"

    def get_app(self):
        return tornado.web.Application(
            panel.get_routes("/"),
            cookie_secret="abc",
            identity_provider=IdentityProvider(token="abc"),
        )

    @respx.mock
    def test_get_project_notebook_v2_success(self):
        respx.get().respond(status_code=200, json=self.success)
        headers = {"Authorization": "token abc", "Cookie": "access_token=xyz", "X-MS-API-Version": "v2"}

        with mock.patch("aext_panels_server.api.ProjectDeploymentMicroserviceAPI._get_headers_and_cookies") as m:
            m.return_value = HeadersAndCookies(headers=headers, cookies={})

            response = self.fetch("/aext_panels_server/projects/test-project/version/test-version/", headers=headers)
            assert response.code == 200
            assert json.loads(response.body) == asdict(
                ServiceResponse(meta=MetaResponse(RequestError(False)), content=self.success)
            )

    @respx.mock
    def test_delete_project_notebook_v2_success(self):
        respx.delete().respond(status_code=204, json=self.success)
        headers = {"Authorization": "token abc", "Cookie": "access_token=xyz", "X-MS-API-Version": "v2"}

        with mock.patch("aext_panels_server.api.ProjectDeploymentMicroserviceAPI._get_headers_and_cookies") as m:
            m.return_value = HeadersAndCookies(headers=headers, cookies={})

            response = self.fetch(
                "/aext_panels_server/projects/test-project/version/test-version/", method="DELETE", headers=headers
            )
            assert response.code == 200
            assert json.loads(response.body) == asdict(
                ServiceResponse(meta=MetaResponse(RequestError(False)), content=self.success)
            )


class TestProjectHandler(AsyncHTTPTestCase):
    def get_app(self):
        return tornado.web.Application(
            panel.get_routes("/"),
            cookie_secret="abc",
            identity_provider=IdentityProvider(token="abc"),
        )

    @respx.mock
    def test_create_project_should_call_log_notebook_data(self):
        respx.get().respond(status_code=200, json={})
        headers = {
            "Authorization": "token abc",
            "Cookie": "access_token=xyz; refresh_token=zyx",
            "X-MS-API-Version": "v1",
        }
        form_data = {
            "path": "backend_lib/panel/tests/assets/test-notebook.ipynb",
            "title": "nb",
            "version": "2024.03.1",
            "description": "",
            "environment": "anaconda-panel-2023.05-py310",
        }
        with mock.patch("aext_panels_server.base_handlers.PABaseHandler._evaluate_feature_flag") as meff:
            meff.return_value = True
            with mock.patch("aext_panels_server.base_handlers.get_server_root_dir") as grd:
                grd.return_value = "./"
                with mock.patch("aext_panels_server.base_handlers.NotebookBaseHandler._get_file_info") as m_gfi:
                    m_gfi.return_value = FileInfo(file_name="test-notebook.ipynb", file_size="120KB")
                    with mock.patch("aext_panels_server.base_handlers.PABaseHandler._log_notebook_data") as m_lbd:
                        self.fetch(
                            "/aext_panels_server/projects",
                            method="POST",
                            headers=headers,
                            body=urllib.parse.urlencode(form_data),
                        )
                        m_lbd.assert_awaited_once()



import os
from enum import Enum


class Environment(str, Enum):
    LOCAL = "local"
    STAGING = "staging"
    PRODUCTION = "production"


class MS_API_Version(str, Enum):
    V1 = "v1"
    V2 = "v2"


# FEATURE FLAGS

# PYTHONANYWHERE
PYTHONANYWHERE_PRODUCTION_SITE = "nb.anaconda.cloud"
PYTHONANYWHERE_PRODUCTION_DOMAIN = "www.pythonanywhere.com"

# PYTHONANYWHERE RESPONSES
WEBAPP_DOMAIN_ALREADY_EXISTS = "domain with this domain name already exists."

# GENERAL
MAX_NUM_WEB_APP_CREATION_ATTEMPTS = 3
MAX_NUM_WEB_APPS = 4

PROJECT_ROOT_PATH = "/var/www/panels/"
PROJECT_LOG_PATH = "/var/log"
ACCESS_LOG = "access_log"
ERROR_LOG = "error_log"
SERVER_LOG = "server_log"

DEFAULT_HTTPX_TIMEOUT_SEC = 3

# PANEL DOMAIN WHERE THE APPS WILL BE PUBLISHED
DEV_DOMAIN_APPS = os.environ.get("PYTHONANYWHERE_DOMAIN", "bp.pa-dev.com")
PROD_DOMAIN_APPS = os.environ.get("PYTHONANYWHERE_DOMAIN", "anacondaapps.cloud")

PANEL_CONDA_EXE = "/opt/conda/condabin/conda"
PANEL_APP_TEMP_DIR = "/var/www/panels/temp"
PANEL_BASE_COMMAND = "/bin/bash -c 'CONDA_EXE={PANEL_CONDA_EXE} {PANEL_CONDA_EXE} project run --project-archive {ARCHIVE_PATH} --directory {PANEL_APP_TEMP_DIR} --external-env {ENVIRONMENT} serve --unix-socket ${{DOMAIN_SOCKET}} --allow-websocket-origin {DOMAIN}  --liveness --log-level trace'"

# KERNELS
PANEL_VALID_KERNELS_PATTERN = "anaconda-panel"  # currently used in production
PANEL_PACKAGE_NAME = "panel"
NBCONVERT_PACKAGE_NAME = "nbconvert"
PANEL_PACKAGE_MINIMUM_VERSION = "1.0.4"

DEV_USE_STAGING_MICROSERVICE = "DEV_USE_STAGING_MICROSERVICE"

# NOTEBOOK METADATA
ALLOWED_CELL_METADATA = ["panel-layout"]
ALLOWED_NB_METADATA = ["panel-cell-order"]


class AppStatus(Enum):
    @classmethod
    def list(cls):
        return list(map(lambda c: c.value, cls))

    RUNNING = "running"
    INACTIVE = "inactive"
    FAILED = "failed"
    PUBLISHED = "published"
    PUBLISHING = "publishing"



from unittest import mock
from uuid import uuid4

import httpx
import pytest
from aext_panels_server.api import project_deployment_ms_api
from aext_panels_server.api_v2 import project_deployment_ms_api_v2
from aext_panels_server.exceptions import CantUpdateProject
from aext_panels_server.request_schemas.notebook_projects_microservice import (
    UpdateProject,
)
from aext_panels_server.services.deployment_microservice import (
    project_microservice_service,
)
from tornado.httputil import HTTPServerRequest

from aext_shared.handler import UserAccessCredentials


@pytest.fixture
def project_id():
    return str(uuid4())


@pytest.fixture
def mock_ms_api_upload_project_archive():
    with mock.patch("aext_panels_server.api.ProjectDeploymentMicroserviceAPI.upload_project_archive") as m:
        m.return_value = httpx.Response(status_code=200)
        yield m


@pytest.fixture
def mock_start_app():
    with mock.patch("aext_panels_server.api.ProjectDeploymentMicroserviceAPI.start_app") as m:
        m.return_value = httpx.Response(
            status_code=200,
            content='{"domains": [{"name": "app.domain.name", "enabled": true}], "webapp_id": 123, "id": "id"}',
        )
        yield m


@pytest.fixture
def mock_ms_api_download_files():
    with mock.patch("aext_panels_server.api.ProjectDeploymentMicroserviceAPI.download_files") as m:
        m.return_value = httpx.Response(
            status_code=200, headers={"content-type": "application/json"}, content='{"archive": 123}'
        )
        yield m


@pytest.fixture
def mock_ms_api_update_project(project_id):
    with mock.patch("aext_panels_server.api.ProjectDeploymentMicroserviceAPI.update_project") as m:
        m.return_value = httpx.Response(
            status_code=200,
            json={
                "id": project_id,
                "title": "Foo",
                "versioning": {
                    "current_version": "2023.08.1",
                    "next_version": "2023.08.2",
                    "all_versions": ["2023.08.1"],
                },
                "notebook_file_name": "file.ipynb",
                "notebook_file_size": "120 Kbytes",
                "permission": "public",
                "environment": "env",
                "description": "New description",
                "last_published_at": None,
                "domains": [],
                "status": "inactive",
            },
        )
        yield m


async def test_update_project(mock_ms_api_upload_project_archive, mock_ms_api_update_project, project_id):
    request_data = UpdateProject(title="New Title", description="New description")
    archive_content = b"ARCHIVE-CONTENT"
    response = await project_microservice_service.update_project(project_id, request_data, archive_content)

    assert response.project_id == project_id
    mock_ms_api_update_project.assert_awaited_once()
    mock_ms_api_upload_project_archive.assert_awaited_once()


async def test_update_project_should_fail_if_upload_fails(
    mock_ms_api_upload_project_archive, mock_ms_api_update_project, project_id
):
    mock_ms_api_upload_project_archive.return_value = httpx.Response(status_code=400)
    request_data = UpdateProject(title="New Title", description="New description")
    archive_content = b"ARCHIVE-CONTENT"
    with pytest.raises(CantUpdateProject):
        await project_microservice_service.update_project(project_id, request_data, archive_content)


async def test_access_credentials_context_manager():
    user_creds = UserAccessCredentials(
        access_token="user-access-token",
        new_access_token="user-new-access-token",
        api_key="user-api-key",
        username="username",
    )
    async with project_microservice_service.access_credentials(user_creds):
        assert project_deployment_ms_api.anaconda_cloud_access_token == "user-access-token"
        assert project_deployment_ms_api_v2.anaconda_cloud_access_token == "user-access-token"
        assert project_deployment_ms_api.anaconda_cloud_username == "username"
        assert project_deployment_ms_api_v2.anaconda_cloud_username == "username"

    assert project_deployment_ms_api.anaconda_cloud_access_token is None
    assert project_deployment_ms_api_v2.anaconda_cloud_access_token is None


async def test_incoming_request_context_manager():
    request = HTTPServerRequest()
    async with project_microservice_service.incoming_request(request):
        assert project_deployment_ms_api.incoming_request == request
        assert project_deployment_ms_api_v2.incoming_request == request

    assert project_deployment_ms_api.incoming_request is None
    assert project_deployment_ms_api_v2.incoming_request is None


async def test_download_project_files(mock_ms_api_download_files):
    response = await project_microservice_service.download_project_files("project-id", "2023.01.1", api_version="v1")
    assert response.meta.error.status is False
    assert response.content == {"archive": 123}


async def test_start_app(mock_start_app):
    response = await project_microservice_service.start_app("app.domain.name", "user-id")
    assert response.meta.error.status is False



import io
import json
from typing import IO, Dict, Optional

import httpx
from aext_panels_server.api import PythonAnywhereAPI
from aext_panels_server.const import PROJECT_ROOT_PATH
from aext_panels_server.logger import custom_logger
from aext_panels_server.schemas import MetaResponse, RequestError, ServiceResponse
from aext_panels_server.utils import call_api


class ProjectIndexService:
    async def create_project_index(self):
        raise NotImplementedError

    async def list_project_index(self):
        raise NotImplementedError

    def get_project_index_file_path(self, include_root_path: bool = False) -> str:
        if include_root_path:
            return f"{PROJECT_ROOT_PATH}.projects/index.json"
        return ".projects/index.json"

    async def get_project_index(self, api: PythonAnywhereAPI) -> ServiceResponse:
        project_index_path = project_index_service.get_project_index_file_path()
        err, response = await call_api(api.download_file(project_index_path), always_return_response=True)

        if not err.status:
            return ServiceResponse(
                meta=MetaResponse(error=RequestError(status=False)),
                content=response.json(),
            )

        if not response:
            return ServiceResponse(
                meta=MetaResponse(
                    error=RequestError(
                        status=True,
                        message="Unexpected error while fetching project index",
                    )
                ),
                content={},
            )

        if err.status:
            err_resp = ServiceResponse(
                meta=MetaResponse(
                    error=RequestError(
                        status=False,
                    )
                ),
                content={},
            )
            if response.status_code == httpx.codes.NOT_FOUND:
                err_resp.set_error_message("Index file not found")
            elif response.status_code == httpx.codes.UNAUTHORIZED:
                err_resp.set_error_message("Failed to fetch index - Unauthorized")
            else:
                err_resp.set_error_message(f"Failed to fetch index - {err.message}")
            return err_resp

    async def get_project_index_or_create(self, api, project_id, project_title) -> io.BytesIO:
        """
        Tries to retrieve and load the index file `index.json`
        which is currently stored in PythonAnywhere server.

        Parameters
        ----------
        api: PythonAnywhereAPI
        project_id: Project's id
        project_title: Project's title

        Returns binary dict containing the project index
        -------
        """
        project_index_response = await self.get_project_index(api)
        project_index = project_index_response.content

        # project_index file doesn't exist then create the json
        if not project_index:
            custom_logger.info("Creating project_index json because it does not exist.")
            project_index = {project_id: {"webapp_id": None, "title": project_title}}
        else:
            # project_id doesn't exist then add the new project
            if project_id not in project_index:
                custom_logger.info("Adding new project to project_index json.")
                new_project = {project_id: {"webapp_id": None, "title": project_title}}
                project_index.update(new_project)
            else:
                # project_id exist then update the project title
                custom_logger.info("Updating project's title on project_index json.")
                project_index[project_id]["title"] = project_title

        return io.BytesIO(json.dumps(project_index).encode("utf-8"))

    async def get_webapp_id(
        self,
        project_id: str,
        project_index: Optional[Dict],
        api: Optional[PythonAnywhereAPI],
    ) -> Optional[int]:
        if not project_index and not api:
            raise Exception("If project index is not given the API is needed in order to fetch it")
        if not project_index and api:
            # Project index was not given, fetching it from PA
            project_index = await self.get_project_index(api)

        if project_index_info := project_index.content.get(project_id):
            if project_index_info:
                return project_index_info.get("webapp_id")
        return None

    async def exists(self, api: PythonAnywhereAPI, project_id: str) -> bool:
        project_index_response = await self.get_project_index(api)
        if not project_index_response.content:
            return False
        return bool(project_index_response.content.get(project_id))

    async def _create_json_bytes(self, data: Dict) -> io.BytesIO:
        return io.BytesIO(json.dumps(data).encode("utf-8"))

    async def save(self, api: PythonAnywhereAPI, project_index_content: IO) -> ServiceResponse:
        err, response = await call_api(
            api.upload_file(
                self.get_project_index_file_path(include_root_path=True),
                project_index_content,
            )
        )
        return ServiceResponse(meta=MetaResponse(error=err), content={})

    async def update_webapp_id(self, api: PythonAnywhereAPI, project_id: str, webapp_id: int) -> ServiceResponse:
        project_index_response = await self.get_project_index(api)
        project_index = project_index_response.content
        try:
            project_index[project_id]["webapp_id"] = webapp_id
        except KeyError:
            custom_logger.error(f"Project {project_id} index not found while updating entry")
            return ServiceResponse(
                meta=MetaResponse(
                    error=RequestError(
                        status=True,
                        message="Project {project_id} index not found while updating entry",
                    )
                ),
                content={},
            )
        return await self.save(api, await self._create_json_bytes(project_index))

    async def delete_project(self, api: PythonAnywhereAPI, project_id: str) -> ServiceResponse:
        project_index_response = await self.get_project_index(api)
        project_index = project_index_response.content
        try:
            del project_index[project_id]
        except KeyError:
            # The key was already removed or does not exist. Failing safe and returning success
            custom_logger.info(f"Project {project_id} index not found while deleting entry")
        return await self.save(api, await self._create_json_bytes(project_index))


project_index_service = ProjectIndexService()



import asyncio
import json
import os
import pathlib
import re
import urllib.parse
from pprint import pformat
from typing import Optional

from aext_panels_server.exceptions import CondaEnvironmentNotFound, FileNotFound
from tornado.escape import json_encode
from tornado.httpclient import AsyncHTTPClient, HTTPRequest

from aext_shared.errors import BackendError, BadRequest
from aext_shared.handler import BackendHandler

from .api import PythonAnywhereAPI
from .logger import custom_logger
from .schemas import FileInfo
from .utils import clean_notebook


class PABaseHandler(BackendHandler):
    def _get_server_root(self):
        # Get the base URL of the Jupyter server
        referer = self.request.headers.get("Referer")
        try:
            if referer:
                parsed = urllib.parse.urlparse(referer)
                path = [el for el in parsed.path.split("/") if el]
                if path:
                    # structure = protocol; domain; jupyterlab; user; user_id
                    return f"{parsed.scheme}://{parsed.netloc}/{path[0]}/{path[1]}/{path[2]}"
        except Exception as ex:
            custom_logger.error(f"Could not get server root from Referer {ex}")

        server_root = f"{self.request.protocol}://{self.request.host}"
        return server_root

    async def _evaluate_feature_flag(self, namespace: str, flag_name: str):
        if namespace not in ["nucleus", "notebooks"]:
            raise ValueError(f"Feature flag namespace {namespace} not known")
        try:
            response = await self.anaconda_cloud_proxy(
                f"feature-flags/evaluate/{namespace}/{flag_name}?feature_flag_default=true",
                method="POST",
            )
            if response.get("remote_status_code") == 200:
                return response.get("remote_data")
        except (BackendError, BadRequest) as ex:
            custom_logger.error(f"Error evaluating feature flag {vars(ex)}")

        return False

    async def _log_notebook_data(self, notebook_path: str, log_event_name: str):
        json_data = {"log": True, "notebook_path": notebook_path, "event_name": log_event_name}
        if await self._evaluate_feature_flag("notebooks", "panels-send-logs-to-snake-eyes"):
            headers = {}
            url = f"{self._get_server_root()}/aext_notebook_info_server/info"
            incoming_cookies = {key: morsel.value for key, morsel in self.request.cookies.items()}
            outbound_cookies = "; ".join([f"{name}={value}" for name, value in incoming_cookies.items()])
            if authorization := self.request.headers.get("Authorization"):
                headers["Authorization"] = authorization
            if xsrf := self.cookies.get("_xsrf"):
                headers["X-Xsrftoken"] = xsrf.value
            if outbound_cookies:
                headers["Cookie"] = outbound_cookies

            http_client = AsyncHTTPClient()
            request = HTTPRequest(
                url,
                method="POST",
                body=json_encode(json_data),
                headers=headers,
            )
            response = await http_client.fetch(request)
            custom_logger.info(f"Log: {response}")

    def initialize(self, **kwargs):
        super().initialize(**kwargs)
        self.notebook_path = kwargs.pop("notebook_path", [])
        self.api = PythonAnywhereAPI(
            csrftoken=self.get_cookie("csrftoken"),
            sessionid=self.get_cookie("sessionid"),
        )
        api_vars_dump = vars(self.api)
        with open("/tmp/app.log", "w") as f:
            f.write(pformat(api_vars_dump))

    def _on_finish(self, future):
        if self.api.session is not None:
            asyncio.ensure_future(self.api.session.close())

    def finish(self, *args, **kwargs):
        status_code = kwargs.pop("status_code", None)
        if status_code:
            self.set_status(status_code)

        if not self._finished:
            future = super().finish(*args, **kwargs)
            future.add_done_callback(self._on_finish)
            return future


class NotebookBaseHandler(PABaseHandler):
    def _get_path(self, path: str) -> str:
        root_dir = get_server_root_dir(self.application.settings)
        rel_path = pathlib.Path(self.notebook_path or path)
        if rel_path.is_absolute():
            notebook_path = str(rel_path)
        else:
            notebook_path = str((root_dir / rel_path).absolute())
        return notebook_path

    async def _get_file_info(self, file_path: str) -> Optional[FileInfo]:
        notebook_path = self._get_path(file_path)
        size_str = "0 Bytes"

        if os.path.isfile(notebook_path):
            # TODO: Return error if file does not exist instead of returning dummy data
            # return 400 with a payload saying that the notebook was not found
            with open(notebook_path, "rb") as f:
                data = f.read()
            data = clean_notebook(data.decode("utf-8")).encode("utf-8")

            file_size = os.path.getsize(notebook_path)
            if file_size < 1024:
                size_str = f"{file_size} Bytes"
            elif file_size < 1024 * 1024:
                size_str = f"{file_size / 1024:.2f} KB"
            elif file_size < 1024 * 1024 * 1024:
                size_str = f"{file_size / (1024 * 1024):.2f} MB"
            else:
                size_str = f"{file_size / (1024 * 1024 * 1024):.2f} GB"

            return FileInfo(file_name=file_path, file_size=size_str)

    async def extract_notebook_environment(self, notebook_file_path: str) -> str:
        """Extract conda environment name from notebook file

        Parameters
        ----------
        notebook_file_path: Notebook file path

        Returns
        -------
        conda environment name
        """
        try:
            with open(notebook_file_path, "r", encoding="utf-8") as f:
                nb = json.load(f)
                kernel = nb.get("metadata", {}).get("kernelspec", {}).get("name")

        except FileNotFound:
            kernel = None

        if not kernel:
            raise CondaEnvironmentNotFound()

        kernelspec = self.kernel_manager.kernel_spec_manager.get_kernel_spec(kernel)
        environment = kernelspec.metadata.get("conda_env_name") if kernelspec else None

        # checks for custom conda environment and strips ".conda-" prefix
        match = re.search(r"^.conda-(.*)", environment)
        if match:
            return match.group(1)

        return environment


def get_server_root_dir(settings):
    if "server_root_dir" in settings:
        # notebook >= 5.0.0 has this in the settings
        root_dir = settings["server_root_dir"]
    else:
        # This copies the logic added in the notebook in
        #  https://github.com/jupyter/notebook/pull/2234
        contents_manager = settings["contents_manager"]
        root_dir = contents_manager.root_dir
    return os.path.expanduser(root_dir)



import httpx
from aext_panels_server.const import PANEL_PACKAGE_MINIMUM_VERSION


def camel_to_snake(s):
    return "".join(["_" + c.lower() if c.isupper() else c for c in s]).lstrip("_")


class JupyterLabPAException(Exception):
    default_status = httpx.codes.INTERNAL_SERVER_ERROR.value
    default_message = "Something went wrong"

    def __init__(self, message=None, status=None, code=None, data=None):
        self.status = status or self.default_status
        self.code = code or camel_to_snake(self.__class__.__name__)
        self.message = message or self.default_message or ""
        self.data = data or {}


class CantRetrieveCondaPackages(JupyterLabPAException):
    default_status = httpx.codes.INTERNAL_SERVER_ERROR.value
    default_message = "Can't retrieve Conda Packages"


class CantGenerateCondaPackages(JupyterLabPAException):
    default_status = httpx.codes.INTERNAL_SERVER_ERROR.value
    default_message = "Can't generate Conda Packages"


class CondaEnvironmentNotFound(JupyterLabPAException):
    default_status = httpx.codes.INTERNAL_SERVER_ERROR.value
    default_message = "Conda environment not found"


class CantCreateZipFile(JupyterLabPAException):
    default_status = httpx.codes.INTERNAL_SERVER_ERROR.value
    default_message = "Can't create project zip file"


class CantRetrieveProjectInfo(JupyterLabPAException):
    default_status = httpx.codes.INTERNAL_SERVER_ERROR.value
    default_message = "Can't retrieve Project info"


class FailedSavingProjectInfo(JupyterLabPAException):
    default_status = httpx.codes.INTERNAL_SERVER_ERROR.value
    default_message = "Can't save Project info"


class FileNotFound(JupyterLabPAException):
    default_status = httpx.codes.NOT_FOUND.value
    default_message = "File not found"


class InvalidProjectVersion(JupyterLabPAException):
    default_status = httpx.codes.INTERNAL_SERVER_ERROR.value
    default_message = "Invalid Project version"


class InvalidPanelVersion(JupyterLabPAException):
    default_status = httpx.codes.BAD_REQUEST.value
    default_message = f"The Panel version is incompatible. Make sure to use a version > {PANEL_PACKAGE_MINIMUM_VERSION}"


class PanelNotInstalled(JupyterLabPAException):
    default_status = httpx.codes.BAD_REQUEST.value
    default_message = "The Panel package was not found."


class NbconvertNotInstalled(JupyterLabPAException):
    default_status = httpx.codes.BAD_REQUEST.value
    default_message = "The nbconvert package was not found."


class CantCreateProject(JupyterLabPAException):
    default_status = httpx.codes.BAD_REQUEST.value
    default_message = "Can't create new project"


class CantUpdateProject(JupyterLabPAException):
    default_status = httpx.codes.BAD_REQUEST.value
    default_message = "Can't update project"


class CantCreateProjectVersion(JupyterLabPAException):
    default_status = httpx.codes.BAD_REQUEST.value
    default_message = "Can't create new project version"


class Unauthorized(JupyterLabPAException):
    default_status = httpx.codes.UNAUTHORIZED.value
    default_message = "Not authenticated"



import json

import pytest
from aext_panels_server.handlers_helpers import create_response


@pytest.mark.parametrize("status", [True, False])
def test_create_response_with_json_response(status):
    response = create_response(status=status, message="Message", content="Content", json_response=True)
    response = json.loads(response)
    assert response["content"] == "Content"
    assert response["meta"]["error"]["message"] == "Message"
    assert response["meta"]["error"]["status"] == status


@pytest.mark.parametrize("status", [True, False])
def test_create_response_with_no_json_response(status):
    response = create_response(status=status, message="Message", content="Content", json_response=False)
    assert response.content == "Content"
    assert response.meta.error.message == "Message"
    assert response.meta.error.status == status



from __future__ import annotations

import asyncio
import calendar
import functools
import json
import os
import re
from dataclasses import asdict
from datetime import datetime
from json.decoder import JSONDecodeError
from types import coroutine
from typing import Any, Coroutine, Dict, List, Optional, Tuple

import httpx
from aext_panels_server.const import (
    ALLOWED_CELL_METADATA,
    ALLOWED_NB_METADATA,
    DEFAULT_HTTPX_TIMEOUT_SEC,
    PYTHONANYWHERE_PRODUCTION_SITE,
    Environment,
)
from aext_panels_server.logger import custom_logger
from aext_panels_server.schemas import RequestError, ServiceResponse


def is_local_development_env():
    return bool(os.getenv("LOCAL_ENV"))


def is_staging_env():
    is_production_site = os.getenv("PYTHONANYWHERE_SITE") != PYTHONANYWHERE_PRODUCTION_SITE
    is_local = is_local_development_env()
    return is_production_site and not is_local


def is_anaconda_toolbox_env():
    # This environment variable are only set in Anaconda Cloud Notebooks
    return not os.getenv("PYTHONANYWHERE_SITE")


def get_environment() -> Environment:
    environment = Environment.PRODUCTION

    if is_anaconda_toolbox_env():
        environment = Environment.PRODUCTION
    elif is_staging_env():
        environment = Environment.STAGING
    elif is_local_development_env():
        environment = Environment.LOCAL

    custom_logger.info(f"****************** ENVIRONMENT {environment} ******************")
    return environment


def _update_value(dct: dict[str, Any], key: str, value: Any):
    if key in dct and dct[key] != value:
        dct[key] = value


def _clean_cell(cell):
    _update_value(cell, "outputs", [])
    _update_value(cell, "execution_count", None)
    clean_metadata = {}
    for key, value in cell.get("metadata", {}).items():
        if key in ALLOWED_CELL_METADATA:
            clean_metadata[key] = value
    _update_value(cell, "metadata", clean_metadata)


def clean_notebook(data: str) -> str:
    """
    Parses notebook as a string and strips out output and metadata
    unless it is allowed.

    Parameters
    ----------
    data: Notebook file as a string

    Returns
    -------
    cleaned_nb: Notebook after stripping output and metadata
    """
    nb = json.loads(data)
    for cell in nb["cells"]:
        _clean_cell(cell)
    metadata = {"language_info": {"name": "python", "pygments_lexer": "ipython3"}}
    for key, value in nb.get("metadata", {}).items():
        if key in ALLOWED_NB_METADATA:
            metadata[key] = value
    _update_value(nb, "metadata", metadata)
    cleaned_nb = json.dumps(nb, ensure_ascii=False)
    return cleaned_nb


def to_json(response: ServiceResponse):
    return json.dumps(asdict(response, dict_factory=lambda x: {k: v for (k, v) in x if v is not None}))


def extract_file_name(file_path, remove_extension: bool = True):
    """
    Remove the directory path, if any, from the file name

    Parameters
    ----------
    file_path: file path, it can be relative or absolute

    Returns
    -------
    file_name: file name
    """
    file_name = file_path.split("/")[-1]

    # Remove the ".ipynb" extension, if any, from the file name
    if remove_extension:
        file_name = re.sub(r"\.ipynb$", "", file_name)

    return file_name


async def call_api(
    api_call: Coroutine, always_return_response: bool = False
) -> Tuple[RequestError, Optional[httpx.Response]]:
    """
    This function receives a coroutine that should be created out of one of the methods in
    PythonAnywhere API class and performs the request call.

    First it will look for any exception related to networking and if that succeed it will then
    check the status code and determine if the request succeeded or not.

    It will return an RequestError with status True in case either an exception is caught
    or if the status code says there was an error with the request,
    otherwise it will return an RequestError object with status False along with the httpx.Reponse object
    got back from PythonAnywhere API.

    Parameters
    ----------
    api_call: Awaitable created out of one of the PythonAnywhere api methods
    always_return_response: Controls if response should be always returned

    Returns
    -------

    """

    def _get_response() -> Optional[httpx.Response]:
        return response if always_return_response else None

    # Fire the HTTP request
    try:
        response = None
        response = await api_call
    except (
        httpx.TimeoutException,
        httpx.ConnectTimeout,
        httpx.ReadTimeout,
        httpx.WriteTimeout,
        httpx.PoolTimeout,
    ) as ex:
        return (
            RequestError(status=True, message=f"{ex.__class__.__name__}"),
            _get_response(),
        )
    except httpx.NetworkError:
        return (
            RequestError(status=True, message="Request Network Error"),
            _get_response(),
        )
    except httpx.ProxyError:
        return (
            RequestError(status=True, message="Request Network Error"),
            _get_response(),
        )
    except Exception:
        return (
            RequestError(status=True, message="Unexpected error when calling api"),
            _get_response(),
        )

    # HTTP request without any networking errors. Checking the status code
    status = response.status_code
    success_range = range(httpx.codes.OK, httpx.codes.IM_USED + 1)
    try:
        if status in success_range:
            return RequestError(status=False, message=""), response
        elif status == httpx.codes.NOT_FOUND:
            return RequestError(status=True, message="Not found"), _get_response()
        elif status >= httpx.codes.BAD_REQUEST and status < httpx.codes.INTERNAL_SERVER_ERROR:
            return (
                RequestError(status=True, message=f"Client error: {response.json()}"),
                _get_response(),
            )
        else:
            return (
                RequestError(status=True, message=f"Server error: {response.json()}"),
                _get_response(),
            )
    except JSONDecodeError:
        return (
            RequestError(status=True, message="Could not decode JSON"),
            _get_response(),
        )
    except Exception:
        return (
            RequestError(status=True, message="An unexpected error occurred while making request"),
            None,
        )


async def run_concurrently(coroutines: Dict[str, coroutine], timeout: int = None):
    """
    Receives a list of coroutines and execute them all concurrently
    returning a dictionary where each key carries the coroutine name and it's
    result as value.

    Args:
        coroutines: dict of coroutines keyed by some unique ID
        timeout: after this timeout the coroutines that have finished will be returned and the
                 remaining ones will keep in pending state

    Returns: Dict with coroutine IDs and it's results

    """
    tasks = []
    for key in coroutines:
        tasks.append(asyncio.tasks.create_task(coroutines[key], name=key))
    task_results, pending = await asyncio.wait(tasks, timeout=timeout, return_when=asyncio.ALL_COMPLETED)
    if pending:
        for task in pending:
            task.cancel()

    return {task.get_name(): task.result() for task in task_results}


def get_vdate():
    return f"{datetime.now().year}.{datetime.now().month:0>2}"


def to_snake_case(names: List):
    """Convert to snake case a list of strings

    Parameters
    ----------
    names: List of strings

    Returns
    -------
    list of string converted to snake case
    """

    normalized_names = []
    for name in names:
        # replace all non-alphanumeric/space with empty space
        step_1 = re.sub("[^A-Za-z0-9 -]+", "", name).lower()
        # replace all non-alphanumeric with underlines
        step_2 = re.sub("[^A-Za-z0-9]+", "-", step_1).lower()

        # replace all non-alphanumeric with underline
        normalized_names.append(step_2)

    return normalized_names


def get_utc_timestamp() -> float:
    """
    Get the current timestamp in UTC
    Returns: timestamp in UTC
    -------
    """

    return calendar.timegm(datetime.utcnow().utctimetuple())


async def read_mocked_response(filename: str, to_json: bool = False):
    """
    Read a mock file
    """
    dirname = os.path.dirname(__file__)
    with open(f"{os.path.join(dirname, f'mocked-responses/{filename}')}", "r") as f:
        if to_json:
            return json.loads(f.read())
        return f.read()


def get_relative_file_path(file_path: str) -> str:
    """
    Return the relative file path
    e.g.:
    /home/user/subdir/filename.ext
    /subdir/filename.ext

    or
    /home/user/filename.ext
    /filename.ext

    Parameters:
    -----------
    file_path: file path to be parsed

    Returns:
    --------
    relative file path parsed
    """

    matches = re.search(r"/([^/]+)/([^/]+)(.*)", file_path)

    if matches:
        relative_file_path = matches.group(3)

        # if there is `/` char then remove it
        if relative_file_path[0] == "/":
            return relative_file_path[1:]

        return matches.group(3)

    return file_path


def escape_filename(cmd: str) -> str:
    """
    Replace expecial chars by underline

    Parameters:
    -----------
    escaped_filename: filename to be escaped

    Returns:
    --------
    filename escaped
    """
    escaped_filename = re.sub(r"[^\w\-_\.]", "_", cmd)  # type: ignore
    return escaped_filename


def handle_network_exceptions(max_retries, default_timeout=DEFAULT_HTTPX_TIMEOUT_SEC):
    def actual_decorator(func):
        @functools.wraps(func)
        async def wrapper(*args, **kwargs):
            try:
                custom_logger.debug(f"Calling {func.__name__}")
                return await func(*args, **kwargs)
            except httpx.TimeoutException:
                timeout = default_timeout * 2
                nonlocal max_retries
                while max_retries:
                    custom_logger.warning(
                        f"Connection timed out while calling {func.__name__}. " f"Retrying with {timeout}s timeout"
                    )
                    max_retries -= 1
                    try:
                        return await func(*args, timeout_sec=timeout, **kwargs)
                    except httpx.TimeoutException:
                        continue

                custom_logger.error(
                    f"Connection timed out while calling {func.__name__}. " f"Number of retries exceeded, giving up"
                )
            except httpx.NetworkError:
                custom_logger.error(
                    f"Network error while calling {func.__name__}. " f"Number of retries exceeded, giving up"
                )

        return wrapper

    return actual_decorator


def get_username() -> str:
    return os.getenv("USERNAME", os.getenv("PYTHONANYWHERE_USER", "unknown"))



import os
from dataclasses import asdict
from typing import IO

import httpx
from aext_panels_server.api import HeadersAndCookies, ProjectDeploymentMicroserviceAPI
from aext_panels_server.const import Environment
from aext_panels_server.request_schemas.notebook_projects_microservice import (
    CreateProject,
    CreateProjectVersion,
)
from aext_panels_server.utils import get_environment, handle_network_exceptions


class ProjectDeploymentMicroserviceAPIV2(ProjectDeploymentMicroserviceAPI):
    NUCLEUS_BASE_URLS = {
        Environment.PRODUCTION: "https://anaconda.cloud/api/panels/v2/",
        Environment.STAGING: "https://nucleus-latest.anacondaconnect.com/api/panels/v2/",
        Environment.LOCAL: os.getenv("DEV_PANELS_MICROSERVICE_ADDRESS_V2", "http://notebook-deployment-api:8000/v2/"),
    }

    def _get_headers_and_cookies(self, **kwargs) -> HeadersAndCookies:
        headers_and_cookies = super()._get_headers_and_cookies()
        headers_and_cookies.headers.update(**kwargs.pop("headers", {}))
        return headers_and_cookies

    @handle_network_exceptions(max_retries=1)
    async def create_project(self, request_data: CreateProject, archive_content: IO, **kwargs) -> httpx.Response:
        # an `app` is an updated term for v1's `project`
        # at some point, we should refactor the code to only use `app`
        endpoint = "apps/"

        data = {**asdict(request_data), "name": request_data.title}
        data.pop("title", None)
        data.pop("description", None)

        archive = {"archive": archive_content}
        return await self._post(endpoint, data=data, files=archive, **kwargs)

    @handle_network_exceptions(max_retries=1)
    async def delete_project(self, project_id: str, **kwargs) -> httpx.Response:
        # an `app` is an updated term for v1's `project`
        # at some point, we should refactor the code to only use `app`
        endpoint = f"apps/{project_id}/"
        return await self._delete(endpoint, **kwargs)

    @handle_network_exceptions(max_retries=1)
    async def download_files(self, project_id: str, project_version: str, **kwargs) -> httpx.Response:
        # an `app` is an updated term for v1's `project`
        # at some point, we should refactor the code to only use `app`
        endpoint = f"apps/{project_id}/download/{project_version}/"
        headers = {"Content-Type": "octet-stream"}  # subject to change
        return await self._get(endpoint, headers=headers, **kwargs)

    @handle_network_exceptions(max_retries=1)
    async def create_project_version(
        self, request_data: CreateProjectVersion, archive_content: IO, **kwargs
    ) -> httpx.Response:
        # an `app` is an updated term for v1's `project`
        # at some point, we should refactor the code to only use `app`
        endpoint = f"apps/{request_data.project_id}/"
        archive = {"archive": archive_content}
        return await self._post(endpoint, data=asdict(request_data), files=archive, **kwargs)

    @handle_network_exceptions(max_retries=1)
    async def get_project_notebook(self, project_id: str, version_id: str, **kwargs) -> httpx.Response:
        # an `app` is an updated term for v1's `project`
        # at some point, we should refactor the code to only use `app`
        endpoint = f"apps/{project_id}/storage/{version_id}/"
        return await self._get(endpoint, **kwargs)

    @handle_network_exceptions(max_retries=1)
    async def delete_project_version(self, project_id: str, version_id: str, **kwargs) -> httpx.Response:
        # an `app` is an updated term for v1's `project`
        # at some point, we should refactor the code to only use `app`
        endpoint = f"apps/{project_id}/version/{version_id}/"
        return await self._delete(endpoint, **kwargs)


project_deployment_ms_api_v2 = ProjectDeploymentMicroserviceAPIV2(get_environment())



from typing import Any, Optional

from .schemas import MetaResponse, RequestError, ServiceResponse
from .utils import to_json


def create_response(
    status: bool = False,
    message: str = "",
    content: Optional[Any] = "",
    json_response=True,
):
    err = RequestError(status=status, message=message)
    response = ServiceResponse(meta=MetaResponse(error=err), content=content)
    if not json_response:
        return response

    return to_json(response)



import os
from pathlib import Path
from unittest import mock

import httpx
import pytest
from aext_panels_server.api import PythonAnywhereAPI
from aext_panels_server.const import AppStatus
from aext_panels_server.schemas import (
    AppVersion,
    MetaResponse,
    PAProjectInfo,
    ProjectInfoData,
    RequestError,
    ServiceResponse,
    VersioningData,
)
from aext_panels_server.services.project import project_service
from aext_panels_server.utils import get_utc_timestamp, get_vdate


@pytest.fixture(autouse=True, scope="session")
def setup_env_vars():
    os.environ["JUPYTERHUB_USER"] = "85c16f17-16e5-4a13-9009-9d9a486b8ebc"
    os.environ["API_TOKEN"] = "123456789"
    os.environ["NUCLEUS_CLOUDFLARE_CLIENT_ID"] = "nucleus-id"
    os.environ["NUCLEUS_CLOUDFLARE_CLIENT_SECRET"] = "nucleus-secret"


@pytest.fixture
def project_name():
    return "dummy-project_name"


@pytest.fixture
def project_description():
    return "Dummy project description"


@pytest.fixture
def project_versions():
    return [
        AppVersion(
            id="0955a586-9358-4628-b69e-58133834ecba",
            version="2023.1.1",
            created_at=str(get_utc_timestamp()),
            updated_at=str(get_utc_timestamp()),
            last_published_at=None,
        ),
        AppVersion(
            id="0955a586-9358-4628-b69e-58133834ecbb",
            version="2023.1.2",
            created_at=str(get_utc_timestamp()),
            updated_at=str(get_utc_timestamp()),
            last_published_at=None,
        ),
        AppVersion(
            id="0955a586-9358-4628-b69e-58133834ecbc",
            version="2023.1.2",
            created_at=str(get_utc_timestamp()),
            updated_at=str(get_utc_timestamp()),
            last_published_at=None,
        ),
    ]


@pytest.fixture
def project_environment():
    return "base"


@pytest.fixture
def project_id():
    return "c9bc616e-ca19-4983-ba1c-7ae0c6f8c190"


@pytest.fixture
def project_path():
    return "/home/user"


@pytest.fixture
def last_published_at():
    return str(get_utc_timestamp())


@pytest.fixture
def notebook_file_name():
    return "my_pretty_nb.ipynb"


@pytest.fixture
def notebook_file_size():
    return "747 Bytes"


@pytest.fixture
def project_info(
    project_versions,
    project_path,
    project_description,
    project_name,
    project_environment,
    project_id,
    last_published_at,
    notebook_file_name,
    notebook_file_size,
):
    return {
        "versions": project_versions,
        "description": project_description,
        "title": project_name,
        "path": project_path,
        "environment": project_environment,
        "project_id": project_id,
        "last_published_at": last_published_at,
        "notebook_file_name": notebook_file_name,
        "notebook_file_size": notebook_file_size,
    }


@pytest.fixture
def project_domains():
    return []


@pytest.fixture
def project_info_data(project_info, project_domains):
    return ProjectInfoData(
        title=project_info["title"],
        project_id=project_info["project_id"],
        description=project_info["description"],
        versioning=VersioningData(
            next_version=project_service._create_or_get_version(get_vdate(), PAProjectInfo(**project_info)),
            current_version=project_info["versions"][-1].version,
            all_versions=project_info["versions"],
        ),
        environment=project_info["environment"],
        notebook_file_name="my_pretty_nb.",
        notebook_file_size=project_info["notebook_file_size"],
        permission="public",
        last_published_at=project_info["last_published_at"],
        status=AppStatus.INACTIVE.value,
        domains=project_domains,
    )


@pytest.fixture
def _status_code():
    return 200


@pytest.fixture
def _json_resp():
    return {}


@pytest.fixture
def _binary_file_resp():
    return b"1234567890"


@pytest.fixture
def api_response(_status_code, _json_resp):
    return httpx.Response(status_code=_status_code, json=_json_resp)


@pytest.fixture
def bin_api_response(_status_code, _binary_file_resp):
    return httpx.Response(status_code=_status_code, content=_binary_file_resp)


@pytest.fixture()
def mock_python_anywhere_api(api_response):
    with mock.patch("aext_panels_server.api.PythonAnywhereAPI._simple_request") as m:
        m.return_value = api_response
        yield m


@pytest.fixture()
def mock_python_anywhere_api_bin(bin_api_response):
    with mock.patch("aext_panels_server.api.PythonAnywhereAPI._simple_request") as m:
        m.return_value = bin_api_response
        yield m


@pytest.fixture(scope="module")
def python_anywhere_api():
    return PythonAnywhereAPI()


@pytest.fixture
def project_index_response():
    return {
        "92a916ea-f0c6-49d6-bd75-16ab2244f134": {
            "webapp_id": 1,
            "title": "My Notebook 1",
        },
        "997c2ec0-9601-4729-a136-b855b88bc30f": {
            "webapp_id": None,
            "title": "My Notebook 2",
        },
    }


@pytest.fixture()
def mock_get_project_index(project_index_response):
    with mock.patch("aext_panels_server.services.project_index.ProjectIndexService.get_project_index") as m:
        m.return_value = ServiceResponse(
            meta=MetaResponse(error=RequestError(status=False)),
            content=project_index_response,
        )
        yield m


@pytest.fixture
def mock_project_index_save():
    with mock.patch("aext_panels_server.services.project_index.ProjectIndexService.save") as m:
        m.return_value = ServiceResponse(meta=MetaResponse(error=RequestError(status=False)), content={})
        yield m


@pytest.fixture()
def mock_services_deployment_status_is_live_enabled():
    with mock.patch("aext_panels_server.services.deployment_status.DeploymentStatus._is_live") as m:
        m.return_value = True
        yield m


@pytest.fixture()
def mock_services_deployment_status_is_live_not_enabled():
    with mock.patch("aext_panels_server.services.deployment_status.DeploymentStatus._is_live") as m:
        m.return_value = False
        yield m


@pytest.fixture
def project_error_log_response():
    with open(f"{Path.cwd()}/tests/assets/error.log", "r") as log_file:
        httpx_response = httpx.Response(200, text=log_file.read())
    return httpx_response


@pytest.fixture
def project_access_log_response():
    with open(f"{Path.cwd()}/tests/assets/access.log", "r") as log_file:
        httpx_response = httpx.Response(200, text=log_file.read())
    return httpx_response



from unittest import mock

import pytest
from aext_panels_server.schemas import (
    ListDeploymentsResponse,
    MetaResponse,
    PAProjectInfo,
    RequestError,
    ServiceResponse,
)
from aext_panels_server.services.panel import panel_service


@pytest.fixture
def project_id_1():
    return "30b9ffb6-fc28-4190-a032-9932c4299148"


@pytest.fixture
def project_id_2():
    return "368bec4f-3899-463d-8ab3-98843d026385"


@pytest.fixture
def project_id_3():
    return "7c48a61c-91f3-405a-800c-0911be9b7393"


@pytest.fixture
def webapp_id_1():
    return 1


@pytest.fixture
def webapp_id_2():
    return 2


@pytest.fixture
def list_deployments_response(webapp_id_1, webapp_id_2):
    content = ListDeploymentsResponse(
        deployments=[
            {
                "id": webapp_id_1,
                "user": "e761da58-732e-4943-af40-5d4caba41bfe",
                "command": "",
                "domains": [
                    {
                        "id": webapp_id_1,
                        "user": "e761da58-732e-4943-af40-5d4caba41bfe",
                        "domain_name": "gdipnixthvcu.nb.anaconda.cloud",
                        "enabled": True,
                        "webapp": webapp_id_1,
                    }
                ],
            },
            {
                "id": webapp_id_2,
                "user": "e761da58-732e-4943-af40-5d4caba41bfe",
                "command": "",
                "domains": [
                    {
                        "id": webapp_id_2,
                        "user": "e761da58-732e-4943-af40-5d4caba41bfe",
                        "domain_name": "jlvfhutudboo.nb.anaconda.cloud",
                        "enabled": True,
                        "webapp": webapp_id_2,
                    }
                ],
            },
        ]
    )
    return ServiceResponse(meta=MetaResponse(error=RequestError(status=False, message="")), content=content)


@pytest.fixture
def project_index_response(project_id_1, project_id_2, project_id_3, webapp_id_1, webapp_id_2):
    return {
        project_id_1: {"webapp_id": webapp_id_1, "title": "My Notebook #1"},
        project_id_2: {"webapp_id": webapp_id_2, "title": "My Notebook #2"},
        project_id_3: {"webapp_id": None, "title": "My Notebook #3"},
    }


@pytest.fixture
def mock_get_project_index(project_index_response):
    with mock.patch("aext_panels_server.services.project_index.ProjectIndexService.get_project_index") as m:
        m.return_value = ServiceResponse(
            meta=MetaResponse(error=RequestError(status=False)),
            content=project_index_response,
        )
        yield m


@pytest.fixture
def mock_get_project_info(python_anywhere_api, project_id_1, project_id_2, project_id_3):
    def args_based_return(*args, **kwargs):
        if args == (python_anywhere_api, project_id_1):
            return PAProjectInfo(
                title="My project #1",
                path="my-notebook-1.ipynb",
                notebook_file_name="my-notebook-1.ipynb",
                notebook_file_size="100 Bytes",
                environment="base",
                project_id=project_id_1,
                description="This is my project #1",
                versions=["2023.04.1"],
            )
        elif args == (python_anywhere_api, project_id_2):
            return PAProjectInfo(
                title="My project #2",
                path="my-notebook-2.ipynb",
                notebook_file_name="my-notebook-2.ipynb",
                notebook_file_size="200 Bytes",
                environment="base",
                project_id=project_id_2,
                description="This is my project #2",
                versions=["2023.04.1", "2023.04.2"],
            )
        elif args == (python_anywhere_api, project_id_3):
            return PAProjectInfo(
                title="My project #3",
                path="my-notebook-3.ipynb",
                notebook_file_name="my-notebook-3.ipynb",
                notebook_file_size="300 Bytes",
                environment="base",
                project_id=project_id_3,
                description="This is my project #3",
                versions=["2023.04.1", "2023.04.2", "2023.04.3"],
            )
        else:
            return Exception("exception occurred")

    with mock.patch("aext_panels_server.services.project.ProjectService.get_project_info") as m:
        m.side_effect = args_based_return
        yield m


@pytest.fixture
def mock_index_service_exists():
    with mock.patch("aext_panels_server.services.project_index.ProjectIndexService.exists") as m:
        m.return_value = True
        yield m


@pytest.fixture
def project_delete_response_content():
    return {}


@pytest.fixture
def project_delete_response_meta():
    return MetaResponse(error=RequestError(status=False, message=""))


@pytest.fixture
def project_delete_response(project_delete_response_meta, project_delete_response_content):
    return ServiceResponse(meta=project_delete_response_meta, content=project_delete_response_content)


@pytest.fixture
def deployment_delete_response_content():
    return {}


@pytest.fixture
def deployment_delete_response_meta():
    return MetaResponse(error=RequestError(status=False, message=""))


@pytest.fixture
def deployment_delete_response(deployment_delete_response_meta, deployment_delete_response_content):
    return ServiceResponse(meta=deployment_delete_response_meta, content=deployment_delete_response_content)


@pytest.fixture
def mock_project_delete(project_delete_response):
    with mock.patch("aext_panels_server.services.project.ProjectService.delete") as m:
        m.return_value = project_delete_response
        yield m


@pytest.fixture()
def mock_utils_call_api_error_log(project_error_log_response):
    with mock.patch("aext_panels_server.services.panel.call_api") as m:
        m.return_value = RequestError(status=False, message=""), project_error_log_response
        yield m


async def test_get_app_logs_should_return_successful_error_log_file(python_anywhere_api, mock_utils_call_api_error_log):
    domain = "great_anaconda.anacondaapps.com"
    response = await panel_service.get_app_log(python_anywhere_api, domain, "error_log")
    assert response.meta.error.status is False
    assert response.content != ""


@pytest.fixture()
def mock_utils_call_api_access_log(project_access_log_response):
    with mock.patch("aext_panels_server.services.panel.call_api") as m:
        m.return_value = RequestError(status=False, message=""), project_access_log_response
        yield m


async def test_get_app_logs_should_return_successful_access_log_file(
    python_anywhere_api, mock_utils_call_api_access_log
):
    domain = "great_anaconda.anacondaapps.com"
    response = await panel_service.get_app_log(python_anywhere_api, domain, "access_log")
    assert response.meta.error.status is False
    assert response.content != ""
